# cython: language_level=3
# cython: profile=True
# Time-stamp: <2025-02-12 18:14:41 Tao Liu>

"""Module Description: For pileup functions.

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""

# ------------------------------------
# python modules
# ------------------------------------
# from MACS3.Utilities.Constants import *
import cython
from cython.cimports.MACS3.Signal.cPosValCalculation import single_end_pileup as c_single_end_pileup
from cython.cimports.MACS3.Signal.cPosValCalculation import write_pv_array_to_bedGraph as c_write_pv_array_to_bedGraph
from cython.cimports.MACS3.Signal.cPosValCalculation import PosVal
from cython.cimports.MACS3.Signal.cPosValCalculation import quick_pileup as c_quick_pileup

# ------------------------------------
# Other modules
# ------------------------------------
import numpy as np
import cython.cimports.numpy as cnp
from cython.cimports.cpython import bool

# ------------------------------------
# C lib
# ------------------------------------
from cython.cimports.libc.stdlib import free

# ------------------------------------
# utility internal functions
# ------------------------------------


@cython.cfunc
@cython.inline
def mean(a: float, b: float) -> float:
    return (a + b) / 2


@cython.cfunc
def clean_up_ndarray(x: cnp.ndarray):
    """ Clean up numpy array in two steps
    """
    i: cython.long

    i = x.shape[0] // 2
    x.resize(100000 if i > 100000 else i, refcheck=False)
    x.resize(0, refcheck=False)
    return


@cython.cfunc
def fix_coordinates(poss: cnp.ndarray, rlength: cython.int) -> cnp.ndarray:
    """Fix the coordinates.
    """
    i: cython.long
    ptr: cython.pointer(cython.int) = cython.cast(cython.pointer(cython.int),
                                                  poss.data)  # pointer

    # fix those negative coordinates
    for i in range(poss.shape[0]):
        if ptr[i] < 0:
            ptr[i] = 0
        else:
            break

    # fix those over-boundary coordinates
    for i in range(poss.shape[0]-1, -1, -1):
        if ptr[i] > rlength:
            ptr[i] = rlength
        else:
            break
    return poss

# ------------------------------------
# functions
# ------------------------------------

# ------------------------------------
# functions for pileup_cmd only
# ------------------------------------

# This function uses pure C code for pileup


@cython.ccall
def pileup_and_write_se(trackI,
                        output_filename: bytes,
                        d: cython.int,
                        scale_factor: cython.float,
                        baseline_value: float = 0.0,
                        directional: bool = True,
                        halfextension: bool = True):
    """Pile up a single-end track and write a bedGraph file.

    This is a thin Cython wrapper that computes pileup using the
    C-accelerated routines in ``cPosValCalculation``.

    Parameters
    ----------
    trackI
        Single-end track object (FWTrackI).
    output_filename
        Output bedGraph path.
    d
        Fragment length estimate.
    scale_factor
        Scalar applied to pileup values.
    baseline_value
        Minimum output value per bin. Default is ``0.0``.
    directional
        If ``True``, extend reads only to 3' direction. If ``False``,
        extend to both sides. Default is ``True``.
    halfextension
        If ``True``, compute shift values from ``d`` using the half-
        extension scheme. Default is ``True``.

    Returns
    -------
    None

    Notes
    -----
    This function is currently only used by the ``macs3 pileup`` command.
    """
    five_shift: cython.long
    three_shift: cython.long
    i: cython.long
    rlength: cython.long
    l_data: cython.long = 0
    chroms: list
    n_chroms: cython.int
    chrom: bytes
    plus_tags: cnp.ndarray
    minus_tags: cnp.ndarray
    chrlengths: dict = trackI.get_rlengths()
    plus_tags_pos: cython.pointer(cython.int)
    minus_tags_pos: cython.pointer(cython.int)
    py_bytes: bytes
    chrom_char: cython.pointer(cython.char)
    _data: cython.pointer(PosVal)

    # This block should be reused to determine the actual shift values
    if directional:
        # only extend to 3' side
        if halfextension:
            five_shift = d//-4  # five shift is used to move cursor towards 5' direction to find the start of fragment
            three_shift = d*3//4  # three shift is used to move cursor towards 3' direction to find the end of fragment
        else:
            five_shift = 0
            three_shift = d
    else:
        # both sides
        if halfextension:
            five_shift = d//4
            three_shift = five_shift
        else:
            five_shift = d//2
            three_shift = d - five_shift
    # end of the block

    chroms = list(chrlengths.keys())
    n_chroms = len(chroms)

    fh = open(output_filename, "w")
    fh.write("")
    fh.close()

    for i in range(n_chroms):
        chrom = chroms[i]
        (plus_tags, minus_tags) = trackI.get_locations_by_chr(chrom)
        rlength = cython.cast(cython.long, chrlengths[chrom])
        plus_tags_pos = cython.cast(cython.p_int, plus_tags.data)
        minus_tags_pos = cython.cast(cython.p_int, minus_tags.data)

        _data = c_single_end_pileup(plus_tags_pos,
                                    plus_tags.shape[0],
                                    minus_tags_pos,
                                    minus_tags.shape[0],
                                    five_shift,
                                    three_shift,
                                    0,
                                    rlength,
                                    scale_factor,
                                    baseline_value,
                                    cython.address(l_data))

        # write
        py_bytes = chrom
        chrom_char = py_bytes
        c_write_pv_array_to_bedGraph(_data,
                                     l_data,
                                     chrom_char,
                                     output_filename,
                                     1)

        # clean
        free(_data)
    return

# function to pileup BAMPE/BEDPE stored in PETrackI object and write to a BEDGraph file
# this function uses c function


@cython.ccall
def pileup_and_write_pe(petrackI,
                        output_filename: bytes,
                        scale_factor: float = 1,
                        baseline_value: float = 0.0):
    """Pile up a paired-end track and write a bedGraph file.

    This is a thin Cython wrapper that computes pileup using the
    C-accelerated routines in ``cPosValCalculation``.

    Parameters
    ----------
    petrackI
        Paired-end track object (PETrackI). Must provide
        ``get_rlengths()`` and ``get_locations_by_chr(chrom)``.
    output_filename
        Output bedGraph path as ``bytes``.
    scale_factor
        Scalar applied to pileup values. Default is ``1``.
    baseline_value
        Minimum output value per bin. Default is ``0.0``.

    Returns
    -------
    None

    Notes
    -----
    This function is currently only used by the ``macs3 pileup`` command.

    Examples
    --------
    .. code-block:: python

        pileup_and_write_pe(
            petrackI,
            b"out.bedGraph",
            scale_factor=1.0,
            baseline_value=0.0,
        )
    """
    chrlengths: dict = petrackI.get_rlengths()
    chroms: list
    n_chroms: cython.int
    i: cython.long
    chrom: bytes
    locs: cnp.ndarray
    locs0: cnp.ndarray
    locs1: cnp.ndarray
    start_pos: cython.pointer(cython.int)
    end_pos: cython.pointer(cython.int)
    py_bytes: bytes
    chrom_char: cython.pointer(cython.char)
    _data: cython.pointer(PosVal)
    l_data: cython.long = 0

    chroms = list(chrlengths.keys())
    n_chroms = len(chroms)

    fh = open(output_filename, "w")
    fh.write("")
    fh.close()

    for i in range(n_chroms):
        chrom = chroms[i]
        locs = petrackI.get_locations_by_chr(chrom)

        locs0 = np.sort(locs['l'])
        locs1 = np.sort(locs['r'])
        start_pos = cython.cast(cython.p_int, locs0.data)  # <int *> locs0.data
        end_pos = cython.cast(cython.p_int, locs1.data)  # <int *> locs1.data

        _data = c_quick_pileup(start_pos,
                               end_pos,
                               locs0.shape[0],
                               scale_factor,
                               baseline_value,
                               cython.address(l_data))

        # write
        py_bytes = chrom
        chrom_char = py_bytes
        c_write_pv_array_to_bedGraph(_data,
                                     l_data,
                                     chrom_char,
                                     output_filename,
                                     1)

        # clean
        free(_data)
    return

# ------------------------------------
# functions for other codes
# ------------------------------------

# general pileup function implemented in cython


@cython.ccall
def se_all_in_one_pileup(plus_tags: cnp.ndarray,
                         minus_tags: cnp.ndarray,
                         five_shift: cython.long,
                         three_shift: cython.long,
                         rlength: cython.int,
                         scale_factor: cython.float,
                         baseline_value: cython.float) -> list:
    """Return pileup given 5' end of fragment at plus or minus strand
    separately, and given shift at both direction to recover a
    fragment. This function is for single end sequencing library
    only. Please directly use 'quick_pileup' function for Pair-end
    library.

    It contains a super-fast and simple algorithm proposed by Jie
    Wang. It will take sorted start positions and end positions, then
    compute pileup values.

    It will return a pileup result in similar structure as
    bedGraph. There are two python arrays:

    [end positions, values] or '[p,v] array' in other description for
    functions within MACS3.

    Two arrays have the same length and can be matched by index. End
    position at index x (p[x]) record continuous value of v[x] from
    p[x-1] to p[x].
    Parameters
    ----------
    plus_tags
        Sorted 5' end positions on the plus strand.
    minus_tags
        Sorted 5' end positions on the minus strand.
    five_shift
        Shift applied toward the 5' direction.
    three_shift
        Shift applied toward the 3' direction.
    rlength
        Chromosome length; coordinates are clipped to ``[0, rlength]``.
    scale_factor
        Scalar applied to pileup values.
    baseline_value
        Minimum output value per bin.

    Returns
    -------
    list
        ``[p, v]`` where ``p`` is an ``int32`` numpy array of end
        positions and ``v`` is a ``float32`` numpy array of values.

    Examples
    --------
    .. code-block:: python

        plus = np.array([10, 50, 100], dtype="i4")
        minus = np.array([30, 80], dtype="i4")
        p, v = se_all_in_one_pileup(
            plus,
        minus,
        five_shift=-25,
        three_shift=75,
        rlength=1000,
        scale_factor=1.0,
        baseline_value=0.0,
        )

    """
    p: cython.int
    pre_p: cython.int
    pileup: cython.int = 0

    i_s: cython.long = 0        # index of start_poss
    i_e: cython.long = 0        # index of end_poss
    i: cython.long
    I: cython.long = 0
    lx: cython.long

    start_poss: cnp.ndarray
    end_poss: cnp.ndarray
    ret_p: cnp.ndarray
    ret_v: cnp.ndarray

    # pointers are used for numpy arrays
    start_poss_ptr: cython.pointer(cython.int)
    end_poss_ptr: cython.pointer(cython.int)
    ret_p_ptr: cython.pointer(cython.int)
    ret_v_ptr: cython.pointer(cython.float)

    start_poss = np.concatenate((plus_tags-five_shift, minus_tags-three_shift))
    end_poss = np.concatenate((plus_tags+three_shift, minus_tags+five_shift))

    # sort
    start_poss.sort()
    end_poss.sort()

    # fix negative coordinations and those extends over end of chromosomes
    start_poss = fix_coordinates(start_poss, rlength)
    end_poss = fix_coordinates(end_poss, rlength)

    lx = start_poss.shape[0]

    start_poss_ptr = cython.cast(cython.pointer(cython.int),
                                 start_poss.data)  # <int32_t *> start_poss.data
    end_poss_ptr = cython.cast(cython.pointer(cython.int),
                               end_poss.data)  # <int32_t *> end_poss.data

    ret_p = np.zeros(2 * lx, dtype="i4")
    ret_v = np.zeros(2 * lx, dtype="f4")

    ret_p_ptr = cython.cast(cython.pointer(cython.int), ret_p.data)
    ret_v_ptr = cython.cast(cython.pointer(cython.float), ret_v.data)

    tmp = [ret_p, ret_v]        # for (endpos,value)

    if start_poss.shape[0] == 0:
        return tmp
    pre_p = min(start_poss_ptr[0], end_poss_ptr[0])

    if pre_p != 0:
        # the first chunk of 0
        ret_p_ptr[0] = pre_p
        ret_v_ptr[0] = max(0, baseline_value)
        ret_p_ptr += 1
        ret_v_ptr += 1
        I += 1

    # pre_v = pileup

    assert start_poss.shape[0] == end_poss.shape[0]
    lx = start_poss.shape[0]

    while i_s < lx and i_e < lx:
        if start_poss_ptr[0] < end_poss_ptr[0]:
            p = start_poss_ptr[0]
            if p != pre_p:
                ret_p_ptr[0] = p
                ret_v_ptr[0] = max(pileup * scale_factor, baseline_value)
                ret_p_ptr += 1
                ret_v_ptr += 1
                I += 1
                pre_p = p
            pileup += 1
            i_s += 1
            start_poss_ptr += 1
        elif start_poss_ptr[0] > end_poss_ptr[0]:
            p = end_poss_ptr[0]
            if p != pre_p:
                ret_p_ptr[0] = p
                ret_v_ptr[0] = max(pileup * scale_factor, baseline_value)
                ret_p_ptr += 1
                ret_v_ptr += 1
                I += 1
                pre_p = p
            pileup -= 1
            i_e += 1
            end_poss_ptr += 1
        else:
            i_s += 1
            i_e += 1
            start_poss_ptr += 1
            end_poss_ptr += 1

    if i_e < lx:
        # add rest of end positions
        for i in range(i_e, lx):
            p = end_poss_ptr[0]
            if p != pre_p:
                ret_p_ptr[0] = p
                ret_v_ptr[0] = max(pileup * scale_factor, baseline_value)
                ret_p_ptr += 1
                ret_v_ptr += 1
                I += 1
                pre_p = p
            pileup -= 1
            end_poss_ptr += 1

    # clean mem
    clean_up_ndarray(start_poss)
    clean_up_ndarray(end_poss)

    # resize
    ret_p.resize(I, refcheck=False)
    ret_v.resize(I, refcheck=False)

    return tmp

# quick pileup implemented in cython


@cython.ccall
def quick_pileup(start_poss: cnp.ndarray,
                 end_poss: cnp.ndarray,
                 scale_factor: cython.float,
                 baseline_value: cython.float) -> list:
    """Compute pileup from fragment start/end positions.

    The algorithm is a fast sweep over sorted start and end positions
    (Jie Wang). It returns a [p, v] array compatible with bedGraph
    semantics.

    Parameters
    ----------
    start_poss
        Sorted fragment start positions.
    end_poss
        Sorted fragment end positions.
    scale_factor
        Scalar applied to pileup values.
    baseline_value
        Minimum output value per bin.

    Returns
    -------
    list
        ``[p, v]`` where ``p`` is an ``int32`` numpy array of end
        positions and ``v`` is a ``float32`` numpy array of values.

    Examples
    --------
    .. code-block:: python
    
        starts = np.array([10, 50, 100], dtype="i4")
        ends = np.array([40, 90, 140], dtype="i4")
        p, v = quick_pileup(starts, ends, scale_factor=1.0, baseline_value=0.0)
    """
    p: cython.int
    pre_p: cython.int
    pileup: cython.int = 0

    i_s: cython.long = 0        # index of plus_tags
    i_e: cython.long = 0        # index of minus_tags
    i: cython.long
    I: cython.long = 0
    ls: cython.long = start_poss.shape[0]
    le: cython.long = end_poss.shape[0]
    l: cython.long = ls + le

    start_poss: cnp.ndarray
    end_poss: cnp.ndarray
    ret_p: cnp.ndarray
    ret_v: cnp.ndarray

    tmp: list

    # pointers are used for numpy arrays
    start_poss_ptr: cython.pointer(cython.int)
    end_poss_ptr: cython.pointer(cython.int)
    ret_p_ptr: cython.pointer(cython.int)
    ret_v_ptr: cython.pointer(cython.float)

    start_poss_ptr = cython.cast(cython.pointer(cython.int),
                                 start_poss.data)  # <int32_t *> start_poss.data
    end_poss_ptr = cython.cast(cython.pointer(cython.int),
                               end_poss.data)  # <int32_t *> end_poss.data

    ret_p = np.zeros(l, dtype="i4")
    ret_v = np.zeros(l, dtype="f4")

    ret_p_ptr = cython.cast(cython.pointer(cython.int), ret_p.data)
    ret_v_ptr = cython.cast(cython.pointer(cython.float), ret_v.data)

    tmp = [ret_p, ret_v]        # for (endpos,value)

    if ls == 0:
        return tmp
    pre_p = min(start_poss_ptr[0], end_poss_ptr[0])

    if pre_p != 0:
        # the first chunk of 0
        ret_p_ptr[0] = pre_p
        ret_v_ptr[0] = max(0, baseline_value)
        ret_p_ptr += 1
        ret_v_ptr += 1
        I += 1

    # pre_v = pileup

    while i_s < ls and i_e < le:
        if start_poss_ptr[0] < end_poss_ptr[0]:
            p = start_poss_ptr[0]
            if p != pre_p:
                ret_p_ptr[0] = p
                ret_v_ptr[0] = max(pileup * scale_factor, baseline_value)
                ret_p_ptr += 1
                ret_v_ptr += 1
                I += 1
                pre_p = p
            pileup += 1
            # if pileup > max_pileup:
            #    max_pileup = pileup
            i_s += 1
            start_poss_ptr += 1
        elif start_poss_ptr[0] > end_poss_ptr[0]:
            p = end_poss_ptr[0]
            if p != pre_p:
                ret_p_ptr[0] = p
                ret_v_ptr[0] = max(pileup * scale_factor, baseline_value)
                ret_p_ptr += 1
                ret_v_ptr += 1
                I += 1
                pre_p = p
            pileup -= 1
            i_e += 1
            end_poss_ptr += 1
        else:
            i_s += 1
            i_e += 1
            start_poss_ptr += 1
            end_poss_ptr += 1

    if i_e < le:
        # add rest of end positions
        for i in range(i_e, le):
            p = end_poss_ptr[0]
            # for p in minus_tags[i_e:]:
            if p != pre_p:
                ret_p_ptr[0] = p
                ret_v_ptr[0] = max(pileup * scale_factor, baseline_value)
                ret_p_ptr += 1
                ret_v_ptr += 1
                I += 1
                pre_p = p
            pileup -= 1
            end_poss_ptr += 1

    ret_p.resize(I, refcheck=False)
    ret_v.resize(I, refcheck=False)

    return tmp

# quick pileup implemented in cython


@cython.ccall
def naive_quick_pileup(sorted_poss: cnp.ndarray, extension: int) -> list:
    """Simple pileup by extending each tag symmetrically.

    Each input position is extended left and right by ``extension``.
    The input must be sorted; no sorting or validation is performed.

    Parameters
    ----------
    sorted_poss
        Sorted tag positions.
    extension
        Extension size applied to both sides.

    Returns
    -------
    list
        ``[p, v]`` where ``p`` is an ``int32`` numpy array of end
        positions and ``v`` is a ``float32`` numpy array of values.
    """
    p: cython.int
    pre_p: cython.int
    pileup: cython.int = 0

    i_s: cython.long = 0  # index of plus_tags
    i_e: cython.long = 0  # index of minus_tags
    i: cython.long
    I: cython.long = 0
    l: cython.long = sorted_poss.shape[0]

    start_poss: cnp.ndarray
    end_poss: cnp.ndarray
    ret_p: cnp.ndarray
    ret_v: cnp.ndarray

    # pointers are used for numpy arrays
    start_poss_ptr: cython.pointer(cython.int)
    end_poss_ptr: cython.pointer(cython.int)
    ret_p_ptr: cython.pointer(cython.int)
    ret_v_ptr: cython.pointer(cython.float)

    start_poss = sorted_poss - extension
    start_poss[start_poss < 0] = 0
    end_poss = sorted_poss + extension

    start_poss_ptr = cython.cast(cython.pointer(cython.int),
                                 start_poss.data)  # <int32_t *> start_poss.data
    end_poss_ptr = cython.cast(cython.pointer(cython.int),
                               end_poss.data)  # <int32_t *> end_poss.data

    ret_p = np.zeros(2*l, dtype="i4")
    ret_v = np.zeros(2*l, dtype="f4")

    ret_p_ptr = cython.cast(cython.pointer(cython.int), ret_p.data)
    ret_v_ptr = cython.cast(cython.pointer(cython.float), ret_v.data)

    if l == 0:
        raise Exception("length is 0")

    pre_p = min(start_poss_ptr[0], end_poss_ptr[0])

    if pre_p != 0:
        # the first chunk of 0
        ret_p_ptr[0] = pre_p
        ret_v_ptr[0] = 0
        ret_p_ptr += 1
        ret_v_ptr += 1
        I += 1

    # pre_v = pileup

    while i_s < l and i_e < l:
        if start_poss_ptr[0] < end_poss_ptr[0]:
            p = start_poss_ptr[0]
            if p != pre_p:
                ret_p_ptr[0] = p
                ret_v_ptr[0] = pileup
                ret_p_ptr += 1
                ret_v_ptr += 1
                I += 1
                pre_p = p
            pileup += 1
            i_s += 1
            start_poss_ptr += 1
        elif start_poss_ptr[0] > end_poss_ptr[0]:
            p = end_poss_ptr[0]
            if p != pre_p:
                ret_p_ptr[0] = p
                ret_v_ptr[0] = pileup
                ret_p_ptr += 1
                ret_v_ptr += 1
                I += 1
                pre_p = p
            pileup -= 1
            i_e += 1
            end_poss_ptr += 1
        else:
            i_s += 1
            i_e += 1
            start_poss_ptr += 1
            end_poss_ptr += 1

    # add rest of end positions
    if i_e < l:
        for i in range(i_e, l):
            p = end_poss_ptr[0]
            if p != pre_p:
                ret_p_ptr[0] = p
                ret_v_ptr[0] = pileup
                ret_p_ptr += 1
                ret_v_ptr += 1
                I += 1
                pre_p = p
            pileup -= 1
            end_poss_ptr += 1

    ret_p.resize(I, refcheck=False)
    ret_v.resize(I, refcheck=False)

    return [ret_p, ret_v]

# general function to compare two pv arrays in cython.


@cython.ccall
def over_two_pv_array(pv_array1: list,
                      pv_array2: list,
                      func: str = "max") -> list:
    """Merge two [p, v] arrays with a pointwise reducer.

    Parameters
    ----------
    pv_array1
        First ``[p, v]`` array, same as output from quick_pileup function.
    pv_array2
        Second ``[p, v]`` array, same as output from quick_pileup function.
    func
        Reducer for overlapping regions. One of ``"max"``, ``"min"``,
        or ``"mean"``. Default is ``"max"``.

    Returns
    -------
    list
        Merged ``[p, v]`` array with the reducer applied to overlap
        regions.

    """
    # pre_p: cython.int

    l1: cython.long
    l2: cython.long
    i1: cython.long = 0
    i2: cython.long = 0
    I: cython.long = 0

    a1_pos: cnp.ndarray
    a2_pos: cnp.ndarray
    ret_pos: cnp.ndarray
    a1_v: cnp.ndarray
    a2_v: cnp.ndarray
    ret_v: cnp.ndarray

    # pointers are used for numpy arrays
    a1_pos_ptr: cython.pointer(cython.int)
    a2_pos_ptr: cython.pointer(cython.int)
    ret_pos_ptr: cython.pointer(cython.int)
    a1_v_ptr: cython.pointer(cython.float)
    a2_v_ptr: cython.pointer(cython.float)
    ret_v_ptr: cython.pointer(cython.float)

    if func == "max":
        f = max
    elif func == "min":
        f = min
    elif func == "mean":
        f = mean
    else:
        raise Exception("Invalid function")

    [a1_pos, a1_v] = pv_array1
    [a2_pos, a2_v] = pv_array2
    ret_pos = np.zeros(a1_pos.shape[0] + a2_pos.shape[0], dtype="i4")
    ret_v = np.zeros(a1_pos.shape[0] + a2_pos.shape[0], dtype="f4")

    a1_pos_ptr = cython.cast(cython.pointer(cython.int), a1_pos.data)
    a1_v_ptr = cython.cast(cython.pointer(cython.float), a1_v.data)
    a2_pos_ptr = cython.cast(cython.pointer(cython.int), a2_pos.data)
    a2_v_ptr = cython.cast(cython.pointer(cython.float), a2_v.data)
    ret_pos_ptr = cython.cast(cython.pointer(cython.int), ret_pos.data)
    ret_v_ptr = cython.cast(cython.pointer(cython.float), ret_v.data)

    l1 = a1_pos.shape[0]
    l2 = a2_pos.shape[0]

    # pre_p = 0
    # remember the previous position in the new bedGraphTrackI object ret

    while i1 < l1 and i2 < l2:
        ret_v_ptr[0] = f(a1_v_ptr[0], a2_v_ptr[0])
        I += 1
        if a1_pos_ptr[0] < a2_pos_ptr[0]:
            # clip a region from pre_p to p1, then set pre_p as p1.
            ret_pos_ptr[0] = a1_pos_ptr[0]
            ret_pos_ptr += 1
            ret_v_ptr += 1
            # pre_p = a1_pos_ptr[0]
            # call for the next p1 and v1
            a1_pos_ptr += 1
            a1_v_ptr += 1
            i1 += 1
        elif a1_pos_ptr[0] > a2_pos_ptr[0]:
            # clip a region from pre_p to p2, then set pre_p as p2.
            ret_pos_ptr[0] = a2_pos_ptr[0]
            ret_pos_ptr += 1
            ret_v_ptr += 1
            # pre_p = a2_pos_ptr[0]
            # call for the next p1 and v1
            a2_pos_ptr += 1
            a2_v_ptr += 1
            i2 += 1
        else:
            # from pre_p to p1 or p2, then set pre_p as p1 or p2.
            ret_pos_ptr[0] = a1_pos_ptr[0]
            ret_pos_ptr += 1
            ret_v_ptr += 1
            # pre_p = a1_pos_ptr[0]
            # call for the next p1, v1, p2, v2.
            a1_pos_ptr += 1
            a1_v_ptr += 1
            i1 += 1
            a2_pos_ptr += 1
            a2_v_ptr += 1
            i2 += 1

    ret_pos.resize(I, refcheck=False)
    ret_v.resize(I, refcheck=False)
    return [ret_pos, ret_v]


@cython.ccall
def naive_call_peaks(pv_array: list, min_v: cython.float,
                     max_v: cython.float = 1e30,
                     max_gap: cython.int = 50,
                     min_length: cython.int = 200):
    """Identify peak summits from a [p, v] array.

    Parameters
    ----------
    pv_array
        ``[p, v]`` array as produced by pileup functions.
    min_v
        Minimum value to be considered part of a peak.
    max_v
        Maximum allowed summit height.
        Default is ``1e30``.
    max_gap
        Maximum gap (in bp) allowed between adjacent peak segments to
        be merged. Default is ``50``.
    min_length
        Minimum peak length (in bp). Default is ``200``.

    Returns
    -------
    list
        List of ``(summit, height)`` tuples.

    """

    pre_p: cython.int
    p: cython.int
    i: cython.int
    x: cython.long           # index used for searching the first peak
    v: cython.double
    peak_content: list     # (pre_p, p, v) for each region in the peak
    ret_peaks: list = []   # returned peak summit and height

    peak_content = []
    (ps, vs) = pv_array
    psn = iter(ps).__next__  # assign the next function to a viable to speed up
    vsn = iter(vs).__next__
    x = 0
    pre_p = 0                   # remember previous position
    while True:
        # find the first region above min_v
        try:         # try to read the first data range for this chrom
            p = psn()
            v = vsn()
        except Exception:
            break
        x += 1                  # index for the next point
        if v > min_v:
            peak_content = [(pre_p, p, v),]
            pre_p = p
            break               # found the first range above min_v
        else:
            pre_p = p

    for i in range(x, len(ps)):
        # continue scan the rest regions
        p = psn()
        v = vsn()
        if v <= min_v:          # not be detected as 'peak'
            pre_p = p
            continue
        # for points above min_v
        # if the gap is allowed
        # gap = pre_p - peak_content[-1][1] or the dist between pre_p and the last p
        if pre_p - peak_content[-1][1] <= max_gap:
            peak_content.append((pre_p, p, v))
        else:
            # when the gap is not allowed, close this peak IF length is larger than min_length
            if peak_content[-1][1] - peak_content[0][0] >= min_length:
                __close_peak(peak_content, ret_peaks, max_v, min_length)
            # reset and start a new peak
            peak_content = [(pre_p, p, v),]
        pre_p = p

    # save the last peak
    if peak_content:
        if peak_content[-1][1] - peak_content[0][0] >= min_length:
            __close_peak(peak_content, ret_peaks, max_v, min_length)
    return ret_peaks


@cython.cfunc
def __close_peak(peak_content,
                 peaks,
                 max_v: cython.float,
                 min_length: cython.int):
    """Internal function to find the summit and height

    If the height is larger than max_v, skip
    """
    tsummit: list = []
    summit: cython.int = 0
    summit_value: cython.float = 0
    tstart: cython.int
    tend: cython.int
    tvalue: cython.float

    for (tstart, tend, tvalue) in peak_content:
        if not summit_value or summit_value < tvalue:
            tsummit = [int((tend+tstart)/2),]
            summit_value = tvalue
        elif summit_value == tvalue:
            tsummit.append(int((tend+tstart)/2))
    summit = tsummit[int((len(tsummit)+1)/2)-1]
    if summit_value < max_v:
        peaks.append((summit, summit_value))
    return



# cython: language_level=3
# cython: profile=True
# Time-stamp: <2025-11-14 16:52:15 Tao Liu>

"""Module for PeakIO IO classes.

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""

# ------------------------------------
# python modules
# ------------------------------------
from itertools import groupby
from operator import itemgetter
import random
import sys

# ------------------------------------
# MACS3 modules
# ------------------------------------

# from MACS3.Utilities.Constants import *

# ------------------------------------
# Other modules
# ------------------------------------
import cython
from cython.cimports.cpython import bool

# ------------------------------------
# constants
# ------------------------------------

# ------------------------------------
# Misc functions
# ------------------------------------


@cython.cfunc
def subpeak_letters(i: cython.int) -> str:
    """Return the alphabetical label for a zero-based subpeak index.

    Args:
        i: Zero-based subpeak index.

    Returns:
        str: Alphabetical label sequence (``a``, ``b``, ..., ``aa``).
    """
    if i < 26:
        return chr(97+i)
    else:
        return subpeak_letters(i // 26) + chr(97 + (i % 26))

# ------------------------------------
# Classes
# ------------------------------------


@cython.cclass
class PeakContent:
    """Represent a narrow peak and its derived statistics.

    Attributes:
        chrom: Chromosome name (bytes).
        start: 0-based inclusive start coordinate.
        end: 0-based exclusive end coordinate.
        length: Peak length in base pairs.
        summit: 0-based summit position.
        score: Peak score reported by MACS3.
        pileup: Tag pileup at the summit.
        pscore: ``-log10(pvalue)`` at the summit.
        fc: Fold enrichment at the summit.
        qscore: ``-log10(qvalue)`` at the summit.
        name: Optional peak identifier.
    """
    chrom: bytes
    start: cython.int
    end: cython.int
    length: cython.int
    summit: cython.int
    score: cython.float
    pileup: cython.float
    pscore: cython.float
    fc: cython.float
    qscore: cython.float
    name: bytes

    def __init__(self,
                 chrom: bytes,
                 start: cython.int,
                 end: cython.int,
                 summit: cython.int,
                 peak_score: cython.float,
                 pileup: cython.float,
                 pscore: cython.float,
                 fold_change: cython.float,
                 qscore: cython.float,
                 name: bytes = b""):
        """Initialise a peak record with positional and score metrics.

        Args:
            chrom: Chromosome name encoded as bytes.
            start: 0-based inclusive start coordinate.
            end: 0-based exclusive end coordinate.
            summit: 0-based summit position.
            peak_score: Peak score reported by MACS3.
            pileup: Tag pileup at the summit.
            pscore: ``-log10(pvalue)`` at the summit.
            fold_change: Fold enrichment at the summit.
            qscore: ``-log10(qvalue)`` at the summit.
            name: Optional peak identifier.
        """
        self.chrom = chrom
        self.start = start
        self.end = end
        self.length = end - start
        self.summit = summit
        self.score = peak_score
        self.pileup = pileup
        self.pscore = pscore
        self.fc = fold_change
        self.qscore = qscore
        self.name = name

    def __getitem__(self, a: str):
        """Return a peak attribute by symbolic key.

        Args:
            a: Attribute name (for example ``"chrom"`` or ``"score"``).

        Returns:
            Any: Value associated with the requested attribute.
        """
        if a == "chrom":
            return self.chrom
        elif a == "start":
            return self.start
        elif a == "end":
            return self.end
        elif a == "length":
            return self.length
        elif a == "summit":
            return self.summit
        elif a == "score":
            return self.score
        elif a == "pileup":
            return self.pileup
        elif a == "pscore":
            return self.pscore
        elif a == "fc":
            return self.fc
        elif a == "qscore":
            return self.qscore
        elif a == "name":
            return self.name

    def __setitem__(self, a: str, v):
        """Assign a new value to a peak attribute.

        Args:
            a: Attribute name to set.
            v: Replacement value.
        """
        if a == "chrom":
            self.chrom = v
        elif a == "start":
            self.start = v
        elif a == "end":
            self.end = v
        elif a == "length":
            self.length = v
        elif a == "summit":
            self.summit = v
        elif a == "score":
            self.score = v
        elif a == "pileup":
            self.pileup = v
        elif a == "pscore":
            self.pscore = v
        elif a == "fc":
            self.fc = v
        elif a == "qscore":
            self.qscore = v
        elif a == "name":
            self.name = v

    def __str__(self):
        """Return a concise textual description of the peak."""
        return "chrom:%s;start:%d;end:%d;score:%f" % (self.chrom,
                                                      self.start,
                                                      self.end,
                                                      self.score)

    def __getstate__(self):
        """Serialise the peak content for pickling.

        Returns:
            tuple: Ordered tuple representing the peak state.
        """
        return (self.chrom,
                self.start,
                self.end,
                self.length,
                self.summit,
                self.score,
                self.pileup,
                self.pscore,
                self.fc,
                self.qscore,
                self.name)

    def __setstate__(self, state):
        """Restore the peak content from a previously serialised tuple.

        Args:
            state: Tuple produced by :meth:`__getstate__`.
        """
        (self.chrom, self.start, self.end, self.length, self.summit,
         self.score, self.pileup, self.pscore, self.fc,
         self.qscore, self.name) = state


@cython.cclass
class PeakIO:
    """Manage in-memory collections of narrow peak intervals.

    Attributes:
        peaks: Dictionary storing peak contents.
        CO_sorted: Whether peaks have been coordinate-sorted.
        total: Total number of peaks.
        name: Collection name used in output.
    """
    # dictionary storing peak contents
    peaks = cython.declare(dict, visibility="public")
    # whether peaks have been sorted by coordinations
    CO_sorted = cython.declare(bool, visibility="public")
    # total number of peaks
    total = cython.declare(cython.long, visibility="public")
    # name
    name = cython.declare(bytes, visibility="public")

    def __init__(self, name: bytes = b"MACS3"):
        """Initialise an empty peak collection."""
        self.peaks = {}
        self.CO_sorted = False
        self.total = 0
        self.name = name

    @cython.ccall
    def add(self,
            chromosome: bytes,
            start: cython.int,
            end: cython.int,
            summit: cython.int = 0,
            peak_score: cython.float = 0,
            pileup: cython.float = 0,
            pscore: cython.float = 0,
            fold_change: cython.float = 0,
            qscore: cython.float = 0,
            name: bytes = b""):
        """Add a peak described by raw coordinates and scores.

        Args:
            chromosome: Chromosome name for the peak.
            start: 0-based inclusive start coordinate.
            end: 0-based exclusive end coordinate.
            summit: 0-based summit position.
            peak_score: Reported peak score.
            pileup: Tag pileup at the summit.
            pscore: ``-log10(pvalue)`` score.
            fold_change: Fold enrichment relative to control.
            qscore: ``-log10(qvalue)`` score.
            name: Optional peak identifier.

        Examples:
            .. code-block:: python

                from MACS3.IO.PeakIO import PeakIO
                peaks = PeakIO()
                peaks.add(b"chr1", 100, 200, summit=150, peak_score=10.0)
        """
        if not self.peaks.has_key(chromosome):
            self.peaks[chromosome] = []
        self.peaks[chromosome].append(PeakContent(chromosome,
                                                  start,
                                                  end,
                                                  summit,
                                                  peak_score,
                                                  pileup,
                                                  pscore,
                                                  fold_change,
                                                  qscore,
                                                  name))
        self.total += 1
        self.CO_sorted = False

    @cython.ccall
    def add_PeakContent(self,
                        chromosome: bytes,
                        peakcontent: PeakContent):
        """Extend the collection with an existing :class:`PeakContent`.

        Args:
            chromosome: Chromosome name under which to store the peak.
            peakcontent: Peak record to append.

        Examples:
            .. code-block:: python

                from MACS3.IO.PeakIO import PeakIO, PeakContent
                peaks = PeakIO()
                peak = PeakContent(b"chr1", 100, 200, 150, 10.0, 5.0, 3.0, 2.0, 1.0)
                peaks.add_PeakContent(b"chr1", peak)
        """
        if not self.peaks.has_key(chromosome):
            self.peaks[chromosome] = []
        self.peaks[chromosome].append(peakcontent)
        self.total += 1
        self.CO_sorted = False

    @cython.ccall
    def get_data_from_chrom(self, chrom: bytes) -> list:
        """Return peaks for ``chrom``, initialising storage if needed.

        Args:
            chrom: Chromosome name to query.

        Returns:
            list: Peaks associated with ``chrom``.

        Examples:
            .. code-block:: python

                peaks = PeakIO()
                chrom_peaks = peaks.get_data_from_chrom(b"chr1")
        """
        if not self.peaks.has_key(chrom):
            self.peaks[chrom] = []
        return self.peaks[chrom]

    def get_chr_names(self) -> set:
        """Return the chromosome names represented in the collection.

        Returns:
            set: Unique chromosome names.

        Examples:
            .. code-block:: python

                peaks = PeakIO()
                names = peaks.get_chr_names()
        """
        return set(self.peaks.keys())

    def sort(self):
        """Sort peaks on each chromosome by ascending start position."""
        chrs: list
        chrom: bytes

        # sort by position
        if self.CO_sorted:
            # if already sorted, quit
            return
        chrs = sorted(list(self.peaks.keys()))
        for chrom in sorted(chrs):
            self.peaks[chrom].sort(key=lambda x: x['start'])
        self.CO_sorted = True
        return

    @cython.ccall
    def randomly_pick(self, n: cython.int, seed: cython.int = 12345):
        """Return a new ``PeakIO`` containing ``n`` randomly sampled peaks.

        Args:
            n: Number of peaks to sample.
            seed: RNG seed to ensure reproducibility.

        Returns:
            PeakIO: Fresh instance populated with sampled peaks.

        Examples:
            .. code-block:: python

                sampled = peaks.randomly_pick(100, seed=42)
        """
        all_pc: list
        chrs: list
        chrom: bytes
        ret_peakio: PeakIO
        p: PeakContent

        assert n > 0
        chrs = sorted(list(self.peaks.keys()))
        all_pc = []
        for chrom in sorted(chrs):
            all_pc.extend(self.peaks[chrom])
        random.seed(seed)
        random.shuffle(all_pc)
        all_pc = all_pc[:n]
        ret_peakio = PeakIO()
        for p in all_pc:
            ret_peakio.add_PeakContent(p["chrom"], p)
        return ret_peakio

    @cython.ccall
    def filter_pscore(self, pscore_cut: cython.double):
        """Filter peaks by minimum ``-log10(pvalue)``.

        Args:
            pscore_cut: Lower bound (inclusive) for ``-log10(pvalue)``.
        """
        chrom: bytes
        new_peaks: dict
        chrs: list

        new_peaks = {}
        chrs = sorted(list(self.peaks.keys()))
        self.total = 0
        for chrom in sorted(chrs):
            new_peaks[chrom] = [p for p in self.peaks[chrom] if p['pscore'] >= pscore_cut]
            self.total += len(new_peaks[chrom])
        self.peaks = new_peaks
        self.CO_sorted = True
        self.sort()

    @cython.ccall
    def filter_qscore(self, qscore_cut: cython.double):
        """Filter peaks by minimum ``-log10(qvalue)``.

        Args:
            qscore_cut: Lower bound (inclusive) for ``-log10(qvalue)``.
        """
        chrom: bytes
        new_peaks: dict
        chrs: list

        new_peaks = {}
        chrs = sorted(list(self.peaks.keys()))
        self.total = 0
        for chrom in sorted(chrs):
            new_peaks[chrom] = [p for p in self.peaks[chrom] if p['qscore'] >= qscore_cut]
            self.total += len(new_peaks[chrom])
        self.peaks = new_peaks
        self.CO_sorted = True
        self.sort()

    @cython.ccall
    def filter_fc(self, fc_low: cython.float, fc_up: cython.float = 0):
        """Filter peaks by fold-change range.

        Args:
            fc_low: Inclusive lower bound on fold change.
            fc_up: Exclusive upper bound; ignored if ``<= 0``.
        """
        chrom: bytes
        new_peaks: dict
        chrs: list

        new_peaks = {}
        chrs = list(self.peaks.keys())
        self.total = 0
        if fc_up > 0 and fc_up > fc_low:
            for chrom in sorted(chrs):
                new_peaks[chrom] = [p for p in self.peaks[chrom] if p['fc'] >= fc_low and p['fc'] < fc_up]
                self.total += len(new_peaks[chrom])
        else:
            for chrom in sorted(chrs):
                new_peaks[chrom] = [p for p in self.peaks[chrom] if p['fc'] >= fc_low]
                self.total += len(new_peaks[chrom])
        self.peaks = new_peaks
        self.CO_sorted = True
        self.sort()

    def filter_score(self, lower_score: cython.float, upper_score: cython.float = 0):
        """Filter peaks by their primary score range.

        Args:
            lower_score: Inclusive lower bound on score.
            upper_score: Exclusive upper bound; if ``<= 0`` the bound is ignored.
        """
        chrom: bytes
        new_peaks: dict
        chrs: list

        new_peaks = {}
        chrs = list(self.peaks.keys())
        self.total = 0
        if upper_score > 0 and upper_score > lower_score:
            for chrom in sorted(chrs):
                new_peaks[chrom] = [p for p in self.peaks[chrom] if p['score'] >= lower_score and p['score'] < upper_score]
                self.total += len(new_peaks[chrom])
        else:
            for chrom in sorted(chrs):
                new_peaks[chrom] = [p for p in self.peaks[chrom] if p['score'] >= lower_score]
                self.total += len(new_peaks[chrom])
        self.peaks = new_peaks
        self.CO_sorted = True
        self.sort()

    def __str__(self):
        """Return a debug-friendly representation of all stored peaks."""
        chrs: list
        n_peak: cython.int
        ret: str
        chrom: bytes
        peaks: list

        ret = ""
        chrs = list(self.peaks.keys())
        n_peak = 0
        for chrom in sorted(chrs):
            for end, group in groupby(self.peaks[chrom], key=itemgetter("end")):
                n_peak += 1
                peaks = list(group)
                if len(peaks) > 1:
                    for i, peak in enumerate(peaks):
                        ret += "chrom:%s\tstart:%d\tend:%d\tname:peak_%d%s\tscore:%.6g\tsummit:%d\n" % (chrom.decode(), peak['start'], peak['end'], n_peak, subpeak_letters(i), peak["score"], peak["summit"])
                else:
                    peak = peaks[0]
                    ret += "chrom:%s\tstart:%d\tend:%d\tname:peak_%d\tscore:%.6g\tsummit:%d\n" % (chrom.decode(), peak['start'], peak['end'], n_peak, peak["score"], peak["summit"])
        return ret

    @cython.cfunc
    def _to_bed(self,
                name_prefix: bytes = b"%s_peak_",
                name: bytes = b"MACS",
                description: bytes = b"%s",
                score_column: str = "score",
                trackline: bool = False,
                print_func=sys.stdout.write):
        """Render peaks in BED5-compatible format via a callback.

        Args:
            name_prefix: Template used to build peak names.
            name: Dataset label interpolated into ``name_prefix``.
            description: Track description for optional header line.
            score_column: Peak attribute to emit as the score field.
            trackline: Whether to emit a UCSC ``track`` header line.
            print_func: Callable accepting pre-formatted lines.
        """
        chrs: list
        n_peak: cython.int
        peakprefix: bytes
        desc: bytes

        chrs = list(self.peaks.keys())
        n_peak = 0
        try:
            peakprefix = name_prefix % name
        except Exception:
            peakprefix = name_prefix
        try:
            desc = description % name
        except Exception:
            desc = description

        if trackline:
            try:
                print_func('track name="%s (peaks)" description="%s" visibility=1\n' % (name.replace(b"\"", b"\\\"").decode(),
                                                                                        desc.replace(b"\"", b"\\\"").decode()))
            except Exception:
                print_func('track name=MACS description=Unknown\n')
        for chrom in sorted(chrs):
            for end, group in groupby(self.peaks[chrom], key=itemgetter("end")):
                n_peak += 1
                peaks = list(group)
                if len(peaks) > 1:
                    for i, peak in enumerate(peaks):
                        print_func("%s\t%d\t%d\t%s%d%s\t%.6g\n" % (chrom.decode(), peak['start'], peak['end'], peakprefix.decode(), n_peak, subpeak_letters(i), peak[score_column]))
                else:
                    peak = peaks[0]
                    print_func("%s\t%d\t%d\t%s%d\t%.6g\n" % (chrom.decode(), peak['start'], peak['end'], peakprefix.decode(), n_peak, peak[score_column]))

    @cython.cfunc
    def _to_summits_bed(self,
                        name_prefix: bytes = b"%s_peak_",
                        name: bytes = b"MACS",
                        description: bytes = b"%s",
                        score_column: str = "score",
                        trackline: bool = False,
                        print_func=sys.stdout.write):
        """Render peak summits in BED5 format via a callback.

        Args:
            name_prefix: Template used to build summit names.
            name: Dataset label interpolated into ``name_prefix``.
            description: Track description for optional header line.
            score_column: Peak attribute to emit as the score field.
            trackline: Whether to emit a UCSC ``track`` header line.
            print_func: Callable accepting pre-formatted lines.
        """
        chrs: list
        n_peak: cython.int
        peakprefix: bytes
        desc: bytes

        chrs = list(self.peaks.keys())
        n_peak = 0
        try:
            peakprefix = name_prefix % name
        except Exception:
            peakprefix = name_prefix
        try:
            desc = description % name
        except Exception:
            desc = description
        if trackline:
            try:
                print_func('track name="%s (summits)" description="%s" visibility=1\n' % (name.replace(b"\"", b"\\\"").decode(),
                                                                                          desc.replace(b"\"", b"\\\"").decode()))
            except Exception:
                print_func('track name=MACS description=Unknown')
        for chrom in sorted(chrs):
            for end, group in groupby(self.peaks[chrom], key=itemgetter("end")):
                n_peak += 1
                peaks = list(group)
                if len(peaks) > 1:
                    for i, peak in enumerate(peaks):
                        summit_p = peak['summit']
                        print_func("%s\t%d\t%d\t%s%d%s\t%.6g\n" % (chrom.decode(), summit_p, summit_p+1, peakprefix.decode(), n_peak, subpeak_letters(i), peak[score_column]))
                else:
                    peak = peaks[0]
                    summit_p = peak['summit']
                    print_func("%s\t%d\t%d\t%s%d\t%.6g\n" % (chrom.decode(), summit_p, summit_p+1, peakprefix.decode(), n_peak, peak[score_column]))

    def tobed(self):
        """Write peaks in BED5 format to ``stdout``.

        The five columns correspond to chromosome, start, end, name, and
        the attribute selected by ``score_column``.
        """
        return self._to_bed(name_prefix=b"%s_peak_", score_column="score", name=self.name, description=b"")

    def to_summits_bed(self):
        """Write peak summits in BED5 format to ``stdout``.

        Each summit is emitted as a one-base interval with the selected score column.
        """
        return self._to_summits_bed(name_prefix=b"%s_peak_", score_column="score", name=self.name, description=b"")

    # these methods are very fast, specifying types is unnecessary
    def write_to_bed(self, fhd,
                     name_prefix: bytes = b"%s_peak_",
                     name: bytes = b"MACS",
                     description: bytes = b"%s",
                     score_column: str = "score",
                     trackline: bool = True):
        """Write peaks to a file handle in BED5 format.

        Args:
            fhd: Writable file-like object.
            name_prefix: Template used to build peak names.
            name: Dataset label interpolated into ``name_prefix``.
            description: Track description for optional header line.
            score_column: Peak attribute to emit as the score field.
            trackline: Whether to emit a UCSC ``track`` header line.

        Examples:
            .. code-block:: python

                with open("peaks.bed", "w") as f:
                    peaks.write_to_bed(f)
        """
        return self._to_bed(name_prefix=name_prefix,
                            name=name,
                            description=description,
                            score_column=score_column,
                            print_func=fhd.write,
                            trackline=trackline)

    def write_to_summit_bed(self, fhd,
                            name_prefix: bytes = b"%s_peak_",
                            name: bytes = b"MACS",
                            description: bytes = b"%s",
                            score_column: str = "score",
                            trackline: bool = False):
        """Write peak summits to a file handle in BED5 format.

        Args:
            fhd: Writable file-like object.
            name_prefix: Template used to build summit names.
            name: Dataset label interpolated into ``name_prefix``.
            description: Track description for optional header line.
            score_column: Peak attribute to emit as the score field.
            trackline: Whether to emit a UCSC ``track`` header line.

        Examples:
            .. code-block:: python

                with open("summits.bed", "w") as f:
                    peaks.write_to_summit_bed(f)
        """
        return self._to_summits_bed(name_prefix=name_prefix, name=name,
                                    description=description, score_column=score_column,
                                    print_func=fhd.write, trackline=trackline)

    def write_to_narrowPeak(self, fhd,
                            name_prefix: bytes = b"%s_peak_",
                            name: bytes = b"MACS",
                            score_column: str = "score",
                            trackline: bool = False):
        """Write peaks in the ENCODE narrowPeak (BED6+4) format.

        Args:
            fhd: Writable file-like object.
            name_prefix: Template used to construct peak identifiers.
            name: Dataset label interpolated into ``name_prefix``.
            score_column: Peak attribute mapped to the narrowPeak score field.
            trackline: Whether to emit a UCSC ``track`` header.

        Examples:
            .. code-block:: python

                with open("peaks.narrowPeak", "w") as f:
                    peaks.write_to_narrowPeak(f)
        """
        n_peak: cython.int
        chrom: bytes
        s: cython.long
        peakname: str

        chrs = list(self.peaks.keys())
        n_peak = 0
        write = fhd.write
        try:
            peakprefix = name_prefix % name
        except Exception:
            peakprefix = name_prefix
        if trackline:
            write("track type=narrowPeak name=\"%s\" description=\"%s\" nextItemButton=on\n" % (name.decode(), name.decode()))
        for chrom in sorted(chrs):
            for end, group in groupby(self.peaks[chrom], key=itemgetter("end")):
                n_peak += 1
                these_peaks = list(group)
                if len(these_peaks) > 1:  # from call-summits
                    for i, peak in enumerate(these_peaks):
                        peakname = "%s%d%s" % (peakprefix.decode(), n_peak, subpeak_letters(i))
                        if peak['summit'] == -1:
                            s = -1
                        else:
                            s = peak['summit'] - peak['start']
                        fhd.write("%s\t%d\t%d\t%s\t%d\t.\t%.6g\t%.6g\t%.6g\t%d\n" %
                                  (chrom.decode(),
                                   peak['start'],
                                   peak['end'],
                                   peakname,
                                   int(10*peak[score_column]),
                                   peak['fc'],
                                   peak['pscore'],
                                   peak['qscore'],
                                   s))
                else:
                    peak = these_peaks[0]
                    peakname = "%s%d" % (peakprefix.decode(), n_peak)
                    if peak['summit'] == -1:
                        s = -1
                    else:
                        s = peak['summit'] - peak['start']
                    fhd.write("%s\t%d\t%d\t%s\t%d\t.\t%.6g\t%.6g\t%.6g\t%d\n" %
                              (chrom.decode(),
                               peak['start'],
                               peak['end'],
                               peakname,
                               int(10*peak[score_column]),
                               peak['fc'],
                               peak['pscore'],
                               peak['qscore'],
                               s))
        return

    @cython.ccall
    def write_to_xls(self, ofhd,
                     name_prefix: bytes = b"%s_peak_",
                     name: bytes = b"MACS"):
        """Export narrow peaks to a tab-delimited ``.xls`` text file.

        Args:
            ofhd: Writable file-like object.
            name_prefix: Template used to build peak identifiers.
            name: Dataset label interpolated into ``name_prefix``.

        Examples:
            .. code-block:: python

                with open("peaks.xls", "w") as f:
                    peaks.write_to_xls(f)
        """
        peakprefix: bytes
        chrs: list
        these_peaks: list
        n_peak: cython.int
        i: cython.int

        write = ofhd.write
        write("\t".join(("chr", "start", "end",  "length",  "abs_summit", "pileup", "-log10(pvalue)", "fold_enrichment", "-log10(qvalue)", "name"))+"\n")

        try:
            peakprefix = name_prefix % name
        except Exception:
            peakprefix = name_prefix

        peaks = self.peaks
        chrs = list(peaks.keys())
        n_peak = 0
        for chrom in sorted(chrs):
            for end, group in groupby(peaks[chrom], key=itemgetter("end")):
                n_peak += 1
                these_peaks = list(group)
                if len(these_peaks) > 1:
                    for i, peak in enumerate(these_peaks):
                        peakname = "%s%d%s" % (peakprefix.decode(), n_peak, subpeak_letters(i))
                        # [start,end,end-start,summit,peak_height,number_tags,pvalue,fold_change,qvalue]
                        write("%s\t%d\t%d\t%d" % (chrom.decode(),
                                                  peak['start']+1,
                                                  peak['end'],
                                                  peak['length']))
                        write("\t%d" % (peak['summit']+1))  # summit position
                        write("\t%.6g" % (round(peak['pileup'], 2)))  # pileup height at summit
                        write("\t%.6g" % (peak['pscore']))  # -log10pvalue at summit
                        write("\t%.6g" % (peak['fc']))  # fold change at summit
                        write("\t%.6g" % (peak['qscore']))  # -log10qvalue at summit
                        write("\t%s" % peakname)
                        write("\n")
                else:
                    peak = these_peaks[0]
                    peakname = "%s%d" % (peakprefix.decode(), n_peak)
                    # [start,end,end-start,summit,peak_height,number_tags,pvalue,fold_change,qvalue]
                    write("%s\t%d\t%d\t%d" % (chrom.decode(),
                                              peak['start']+1,
                                              peak['end'],
                                              peak['length']))
                    write("\t%d" % (peak['summit']+1))  # summit position
                    write("\t%.6g" % (round(peak['pileup'], 2)))  # pileup height at summit
                    write("\t%.6g" % (peak['pscore']))  # -log10pvalue at summit
                    write("\t%.6g" % (peak['fc']))  # fold change at summit
                    write("\t%.6g" % (peak['qscore']))  # -log10qvalue at summit
                    write("\t%s" % peakname)
                    write("\n")
        return

    @cython.ccall
    def exclude(self, peaksio2: object):
        """Remove peaks that overlap any entry in ``peaksio2``.

        Args:
            peaksio2: Another :class:`PeakIO` instance providing exclusion regions.
        """
        peaks1: dict
        peaks2: dict
        chrs1: list
        chrs2: list
        k: bytes
        ret_peaks: dict
        overlap_found: bool
        r1: PeakContent
        r2: PeakContent
        n_rl1: cython.long
        n_rl2: cython.long

        self.sort()
        peaks1 = self.peaks
        self.total = 0
        assert isinstance(peaksio2, PeakIO)
        peaksio2.sort()
        peaks2 = peaksio2.peaks

        ret_peaks = dict()
        chrs1 = list(peaks1.keys())
        chrs2 = list(peaks2.keys())
        for k in chrs1:
            #print(f"chromosome {k}")
            if not chrs2.count(k):
                # no such chromosome in peaks1, then don't touch the peaks in this chromosome
                ret_peaks[k] = peaks1[k]
                continue
            ret_peaks[k] = []
            n_rl1 = len(peaks1[k])
            n_rl2 = len(peaks2[k])
            rl1_k = iter(peaks1[k]).__next__
            rl2_k = iter(peaks2[k]).__next__
            overlap_found = False
            r1 = rl1_k()
            n_rl1 -= 1
            r2 = rl2_k()
            n_rl2 -= 1
            while (True):
                # we do this until there is no r1 or r2 left.
                if r2["start"] < r1["end"] and r1["start"] < r2["end"]:
                    # since we found an overlap, r1 will be skipped/excluded
                    # and move to the next r1
                    overlap_found = True
                    # print(f"found overlap of {r1['start']} {r1['end']} and {r2['start']} {r2['end']}, move to the next r1")
                    n_rl1 -= 1
                    if n_rl1 >= 0:
                        r1 = rl1_k()
                        # print(f"move to next r1 {r1['start']} {r1['end']}")
                        overlap_found = False
                        continue
                    else:
                        break
                if r1["end"] < r2["end"]:
                    # print(f"now we need to move r1 {r1['start']} {r1['end']}")
                    # in this case, we need to move to the next r1,
                    # we will check if overlap_found is true, if not, we put r1 in a new dict
                    if not overlap_found:
                        # print(f"we add this r1 {r1['start']} {r1['end']} to list")
                        ret_peaks[k].append(r1)
                    n_rl1 -= 1
                    if n_rl1 >= 0:
                        r1 = rl1_k()
                        # print(f"move to next r1 {r1['start']} {r1['end']}")
                        overlap_found = False
                    else:
                        # no more r1 left
                        break
                else:
                    # in this case, we need to move the next r2
                    if n_rl2:
                        r2 = rl2_k()
                        n_rl2 -= 1
                        # print(f"move to next r2 {r2['start']} {r2['end']}")
                    else:
                        # no more r2 left
                        break
            # add the rest of r1
            # print( f"n_rl1: {n_rl1} n_rl2:{n_rl2} last overlap_found is {overlap_found}" )
            # if overlap_found:
            #    n_rl1 -= 1
            if n_rl1 >= 0:
                ret_peaks[k].extend(peaks1[k][-n_rl1-1:])

        for k in ret_peaks.keys():
            self.total += len(ret_peaks[k])

        self.peaks = ret_peaks
        self.CO_sorted = True
        self.sort()
        return

    @cython.ccall
    def read_from_xls(self, ofhd):
        """Load peak records from a MACS3 ``.xls`` tab-delimited report.

        Args:
            ofhd: Readable file-like object positioned at the beginning of the report.
        """
        line: bytes = b''
        chrom: bytes = b''
        start: cython.int
        end: cython.int
        length: cython.int
        summit: cython.int
        pileup: cython.float
        pscore: cython.float
        fc: cython.float
        qscore: cython.float
        fields: list

        while True:
            if not (line.startswith('#') or line.strip() == ''):
                break
            line = ofhd.readline()

        # sanity check
        columns = line.rstrip().split('\t')
        for a, b in zip(columns, ("chr", "start", "end",  "length", "abs_summit",
                                  "pileup", "-log10(pvalue)", "fold_enrichment",
                                  "-log10(qvalue)", "name")):
            if not a == b:
                raise NotImplementedError('column %s not recognized', a)

        add = self.add
        split = str.split
        rstrip = str.rstrip
        for i, line in enumerate(ofhd.readlines()):
            fields = split(line, '\t')
            chrom = fields[0].encode()
            start = int(fields[1]) - 1
            end = int(fields[2])
            length = int(fields[3])
            if end - start != length:
                raise UserWarning('Malformed peak at line %d:\n%s' % (i, line))
            summit = int(fields[4]) - 1
            pileup = float(fields[5])
            pscore = float(fields[6])
            fc = float(fields[7])
            qscore = float(fields[8])
            peakname = rstrip(fields[9])
            add(chrom, start, end, summit, qscore, pileup, pscore, fc, qscore,
                peakname)


@cython.cclass
class RegionIO:
    """Helper for storing and manipulating simple genomic regions.

    Attributes:
        regions: Mapping of chromosome name to a list of ``(start, end)`` tuples.
        __flag_sorted: Whether per-chromosome regions are sorted.
    """
    regions: dict
    __flag_sorted: bool

    def __init__(self):
        """Initialise an empty region collection."""
        self.regions = {}
        self.__flag_sorted = False

    @cython.ccall
    def add_loc(self, chrom: bytes, start: cython.int, end: cython.int):
        """Append a new ``(start, end)`` interval for ``chrom``.

        Examples:
            .. code-block:: python

                regions = RegionIO()
                regions.add_loc(b"chr1", 100, 200)
        """
        if self.regions.has_key(chrom):
            self.regions[chrom].append((start, end))
        else:
            self.regions[chrom] = [(start, end), ]
        self.__flag_sorted = False
        return

    @cython.ccall
    def sort(self):
        """Sort regions for each chromosome by their start coordinate. """
        chrom: bytes

        for chrom in sorted(list(self.regions.keys())):
            self.regions[chrom].sort()
        self.__flag_sorted = True

    @cython.ccall
    def get_chr_names(self) -> set:
        """Return chromosome names present in the region set."""
        return set(sorted(self.regions.keys()))

    @cython.ccall
    def merge_overlap(self):
        """Merge overlapping intervals within each chromosome."""
        chrom: bytes
        s_new_region: cython.int
        e_new_region: cython.int
        i: cython.int
        regions: dict
        new_regions: dict
        chrs: list
        regions_chr: list
        prev_region: tuple

        if not self.__flag_sorted:
            self.sort()
        regions = self.regions
        new_regions = {}
        chrs = sorted(list(regions.keys()))
        for i in range(len(chrs)):
            chrom = chrs[i]
            new_regions[chrom] = []
            n_append = new_regions[chrom].append
            prev_region = None
            regions_chr = regions[chrom]
            for i in range(len(regions_chr)):
                if not prev_region:
                    prev_region = regions_chr[i]
                    continue
                else:
                    if regions_chr[i][0] <= prev_region[1]:
                        s_new_region = prev_region[0]
                        e_new_region = regions_chr[i][1]
                        prev_region = (s_new_region, e_new_region)
                    else:
                        n_append(prev_region)
                        prev_region = regions_chr[i]
            if prev_region:
                n_append(prev_region)
        self.regions = new_regions
        self.sort()
        return

    @cython.ccall
    def write_to_bed(self, fhd):
        """Emit regions in BED format to the provided file-like object.

        Examples:
            .. code-block:: python

                with open("regions.bed", "w") as f:
                    regions.write_to_bed(f)
        """
        i: cython.int
        chrom: bytes
        chrs: list
        region: tuple

        chrs = sorted(list(self.regions.keys()))
        for i in range(len(chrs)):
            chrom = chrs[i]
            for region in self.regions[chrom]:
                fhd.write("%s\t%d\t%d\n" % (chrom.decode(),
                                            region[0],
                                            region[1]))


@cython.cclass
class BroadPeakContent:
    """Container for broad peak metadata used in broadPeak format."""
    start: cython.int
    end: cython.int
    length: cython.int
    score: cython.float
    thickStart: bytes
    thickEnd: bytes
    blockNum: cython.int
    blockSizes: bytes
    blockStarts: bytes
    pileup: cython.float
    pscore: cython.float
    fc: cython.float
    qscore: cython.float
    name: bytes

    def __init__(self,
                 start: cython.int,
                 end: cython.int,
                 score: cython.float,
                 thickStart: bytes,
                 thickEnd: bytes,
                 blockNum: cython.int,
                 blockSizes: bytes,
                 blockStarts: bytes,
                 pileup: cython.float,
                 pscore: cython.float,
                 fold_change: cython.float,
                 qscore: cython.float,
                 name: bytes = b"MACS3"):
        """Initialise a broad peak record with block structure and scores."""
        self.start = start
        self.end = end
        self.score = score
        self.thickStart = thickStart
        self.thickEnd = thickEnd
        self.blockNum = blockNum
        self.blockSizes = blockSizes
        self.blockStarts = blockStarts
        self.length = end - start
        self.pileup = pileup
        self.pscore = pscore
        self.fc = fold_change
        self.qscore = qscore
        self.name = name

    def __getitem__(self, a):
        """Provide dict-like read access to stored attributes."""
        if a == "start":
            return self.start
        elif a == "end":
            return self.end
        elif a == "length":
            return self.length
        elif a == "score":
            return self.score
        elif a == "thickStart":
            return self.thickStart
        elif a == "thickEnd":
            return self.thickEnd
        elif a == "blockNum":
            return self.blockNum
        elif a == "blockSizes":
            return self.blockSizes
        elif a == "blockStarts":
            return self.blockStarts
        elif a == "pileup":
            return self.pileup
        elif a == "pscore":
            return self.pscore
        elif a == "fc":
            return self.fc
        elif a == "qscore":
            return self.qscore
        elif a == "name":
            return self.name

    def __str__(self):
        """Return a compact summary describing the broad peak."""
        return "start:%d;end:%d;score:%f" % (self.start, self.end, self.score)


@cython.cclass
class BroadPeakIO:
    """IO for broad peak information.

    Attributes:
        peaks: Mapping of chromosome name to :class:`BroadPeakContent` list.
    """
    peaks = cython.declare(dict, visibility="public")

    def __init__(self):
        """Create an empty container for :class:`BroadPeakContent` objects."""
        self.peaks = {}

    @cython.ccall
    def add(self,
            chromosome: bytes,
            start: cython.int,
            end: cython.int,
            score: cython.float = 0.0,
            thickStart: bytes = b".",
            thickEnd: bytes = b".",
            blockNum: cython.int = 0,
            blockSizes: bytes = b".",
            blockStarts: bytes = b".",
            pileup: cython.float = 0,
            pscore: cython.float = 0,
            fold_change: cython.float = 0,
            qscore: cython.float = 0,
            name: bytes = b"NA"):
        """Append a :class:`BroadPeakContent` record.

        Args:
            chromosome: Chromosome name for the region.
            start: 0-based inclusive start coordinate.
            end: 0-based exclusive end coordinate.
            score: Average score across blocks.
            thickStart: Start of the high-enrichment segment or ``b'.'``.
            thickEnd: End of the high-enrichment segment or ``b'.'``.
            blockNum: Number of sub-blocks composing the region.
            blockSizes: Comma-separated block sizes as bytes.
            blockStarts: Comma-separated block starts as bytes.
            pileup: Median pileup within the region.
            pscore: Median ``-log10(pvalue)``.
            fold_change: Median fold-change value.
            qscore: Median ``-log10(qvalue)``.
            name: Optional region identifier.

        Examples:
            .. code-block:: python

                from MACS3.IO.PeakIO import BroadPeakIO
                peaks = BroadPeakIO()
                peaks.add(b"chr1", 100, 500, score=10.0, blockNum=1,
                          blockSizes=b"400", blockStarts=b"0",...)
        """
        if not self.peaks.has_key(chromosome):
            self.peaks[chromosome] = []
        self.peaks[chromosome].append(BroadPeakContent(start,
                                                       end,
                                                       score,
                                                       thickStart,
                                                       thickEnd,
                                                       blockNum,
                                                       blockSizes,
                                                       blockStarts,
                                                       pileup,
                                                       pscore,
                                                       fold_change,
                                                       qscore,
                                                       name))

    @cython.ccall
    def filter_pscore(self, pscore_cut: cython.float):
        """Retain broad peaks with ``-log10(pvalue)`` ≥ ``pscore_cut``.

        Args:
            pscore_cut: Inclusive lower bound for ``-log10(pvalue)``."""
        chrom: bytes
        peaks: dict
        new_peaks: dict
        chrs: list

        peaks = self.peaks
        new_peaks = {}
        chrs = list(peaks.keys())

        for chrom in sorted(chrs):
            new_peaks[chrom] = [p for p in peaks[chrom] if p['pscore'] >= pscore_cut]
        self.peaks = new_peaks

    @cython.ccall
    def filter_qscore(self, qscore_cut: cython.float):
        """Retain broad peaks with ``-log10(qvalue)`` ≥ ``qscore_cut``.

        Args:
            qscore_cut: Inclusive lower bound for ``-log10(qvalue)``."""
        chrom: bytes
        peaks: dict
        new_peaks: dict
        chrs: list

        peaks = self.peaks
        new_peaks = {}
        chrs = list(peaks.keys())

        for chrom in sorted(chrs):
            new_peaks[chrom] = [p for p in peaks[chrom] if p['qscore'] >= qscore_cut]
        self.peaks = new_peaks

    @cython.ccall
    def filter_fc(self, fc_low: float, fc_up: float = -1):
        """Filter broad peaks by fold-change range.

        Args:
            fc_low: Inclusive lower bound on fold change.
            fc_up: Exclusive upper bound; ignored when negative.
        """
        chrom: bytes
        peaks: dict
        new_peaks: dict
        chrs: list

        peaks = self.peaks
        new_peaks = {}
        chrs = list(peaks.keys())
        if fc_up >= 0:
            for chrom in sorted(chrs):
                new_peaks[chrom] = [p for p in peaks[chrom] if p['fc'] >= fc_low and p['fc'] < fc_up]
        else:
            for chrom in sorted(chrs):
                new_peaks[chrom] = [p for p in peaks[chrom] if p['fc'] >= fc_low]
        self.peaks = new_peaks

    @cython.ccall
    def total(self):
        """Return the total number of broad peaks currently stored.

        Returns:
            int: Number of broad peaks."""
        chrom: bytes
        peaks: dict
        chrs: list
        x: cython.long = 0

        peaks = self.peaks
        chrs = list(peaks.keys())
        for chrom in sorted(chrs):
            x += len(peaks[chrom])
        return x

    @cython.ccall
    def write_to_gappedPeak(self, fhd,
                            name_prefix: bytes = b"peak_",
                            name: bytes = b'peak',
                            description: bytes = b"%s",
                            score_column: str = "score",
                            trackline: bool = True):
        """Write broad peaks in gappedPeak (BED12+3) format.

        Args:
            fhd: Writable file-like object.
            name_prefix: Template used to construct peak identifiers.
            name: Dataset label interpolated into ``name_prefix``.
            description: Track description for the optional header.
            score_column: Peak attribute mapped to the score column.
            trackline: Whether to emit a UCSC ``track`` header.

        Examples:
            .. code-block:: python

                with open("broad.gappedPeak", "w") as f:
                    peaks.write_to_gappedPeak(f)
        """
        chrs: list
        n_peak: cython.int = 0
        peak: BroadPeakContent
        desc: bytes
        peakprefix: bytes
        chrom: bytes

        chrs = list(self.peaks.keys())
        try:
            peakprefix = name_prefix % name
        except Exception:
            peakprefix = name_prefix
        try:
            desc = description % name
        except Exception:
            desc = description
        if trackline:
            fhd.write("track name=\"%s\" description=\"%s\" type=gappedPeak nextItemButton=on\n" % (name.decode(), desc.decode()))
        for chrom in sorted(chrs):
            for peak in self.peaks[chrom]:
                n_peak += 1
                if peak["thickStart"] != b".":
                    fhd.write("%s\t%d\t%d\t%s%d\t%d\t.\t0\t0\t0\t%d\t%s\t%s\t%.6g\t%.6g\t%.6g\n" %
                              (chrom.decode(),
                               peak["start"],
                               peak["end"],
                               peakprefix.decode(),
                               n_peak,
                               int(10*peak[score_column]),
                               peak["blockNum"],
                               peak["blockSizes"].decode(),
                               peak["blockStarts"].decode(),
                               peak['fc'],
                               peak['pscore'],
                               peak['qscore']))

    @cython.ccall
    def write_to_Bed12(self, fhd,
                       name_prefix: bytes = b"peak_",
                       name: bytes = b'peak',
                       description: bytes = b"%s",
                       score_column: str = "score",
                       trackline: bool = True):
        """Write broad peaks in BED12 format.

        Args:
            fhd: Writable file-like object.
            name_prefix: Template used to construct peak identifiers.
            name: Dataset label interpolated into ``name_prefix``.
            description: Track description for the optional header.
            score_column: Peak attribute mapped to the score column.
            trackline: Whether to emit a UCSC ``track`` header.

        Examples:
            .. code-block:: python

                with open("broad.bed12", "w") as f:
                    peaks.write_to_Bed12(f)
        """
        chrs: list
        n_peak: cython.int = 0
        peakprefix: bytes
        peak: BroadPeakContent
        desc: bytes
        peakprefix: bytes
        chrom: bytes

        chrs = list(self.peaks.keys())
        try:
            peakprefix = name_prefix % name
        except Exception:
            peakprefix = name_prefix
        try:
            desc = description % name
        except Exception:
            desc = description
        if trackline:
            fhd.write("track name=\"%s\" description=\"%s\" type=bed nextItemButton=on\n" % (name.decode(), desc.decode()))
        for chrom in sorted(chrs):
            for peak in self.peaks[chrom]:
                n_peak += 1
                if peak["thickStart"] == b".":
                    # this will violate gappedPeak format, since it's a complement like broadPeak line.
                    fhd.write("%s\t%d\t%d\t%s%d\t%d\t.\n" %
                              (chrom.decode(),
                               peak["start"],
                               peak["end"],
                               peakprefix.decode(),
                               n_peak,
                               int(10*peak[score_column])))
                else:
                    fhd.write("%s\t%d\t%d\t%s%d\t%d\t.\t%s\t%s\t0\t%d\t%s\t%s\n" %
                              (chrom.decode(),
                               peak["start"],
                               peak["end"],
                               peakprefix.decode(),
                               n_peak,
                               int(10*peak[score_column]),
                               peak["thickStart"].decode(),
                               peak["thickEnd"].decode(),
                               peak["blockNum"],
                               peak["blockSizes"].decode(),
                               peak["blockStarts"].decode()))

    @cython.ccall
    def write_to_broadPeak(self, fhd,
                           name_prefix: bytes = b"peak_",
                           name: bytes = b'peak',
                           description: bytes = b"%s",
                           score_column: str = "score",
                           trackline: bool = True):
        """Write broad peaks in the ENCODE broadPeak (BED6+3) format.

        Args:
            fhd: Writable file-like object.
            name_prefix: Template used to construct peak identifiers.
            name: Dataset label interpolated into ``name_prefix``.
            description: Track description for the optional header.
            score_column: Peak attribute mapped to the score column.
            trackline: Whether to emit a UCSC ``track`` header.

        Examples:
            .. code-block:: python

                with open("broad.broadPeak", "w") as f:
                    peaks.write_to_broadPeak(f)
        """
        chrs: list
        n_peak: cython.int = 0
        peakprefix: bytes
        peak: BroadPeakContent
        peakprefix: bytes
        chrom: bytes
        peakname: str

        chrs = list(self.peaks.keys())
        write = fhd.write
        try:
            peakprefix = name_prefix % name
        except Exception:
            peakprefix = name_prefix
        if trackline:
            write("track type=broadPeak name=\"%s\" description=\"%s\" nextItemButton=on\n" % (name.decode(), name.decode()))
        for chrom in sorted(chrs):
            for end, group in groupby(self.peaks[chrom], key=itemgetter("end")):
                n_peak += 1
                these_peaks = list(group)
                peak = these_peaks[0]
                peakname = "%s%d" % (peakprefix.decode(), n_peak)
                fhd.write("%s\t%d\t%d\t%s\t%d\t.\t%.6g\t%.6g\t%.6g\n" %
                          (chrom.decode(),
                           peak['start'],
                           peak['end'],
                           peakname,
                           int(10*peak[score_column]),
                           peak['fc'],
                           peak['pscore'],
                           peak['qscore']))
        return

    @cython.ccall
    def write_to_xls(self, ofhd,
                     name_prefix: bytes = b"%s_peak_",
                     name: bytes = b"MACS"):
        """Export broad peaks to a tab-delimited ``.xls`` text file.

        Args:
            ofhd: Writable file-like object.
            name_prefix: Template used to build peak identifiers.
            name: Dataset label interpolated into ``name_prefix``.

        Examples:
            .. code-block:: python

                with open("broad.xls", "w") as f:
                    peaks.write_to_xls(f)
        """
        chrom: bytes
        chrs: list
        peakprefix: bytes
        peaks: dict
        these_peaks: list
        peak: BroadPeakContent
        peakname: str

        write = ofhd.write
        write("\t".join(("chr", "start", "end",  "length",  "pileup", "-log10(pvalue)", "fold_enrichment", "-log10(qvalue)", "name"))+"\n")

        try:
            peakprefix = name_prefix % name
        except Exception:
            peakprefix = name_prefix

        peaks = self.peaks
        chrs = list(peaks.keys())
        n_peak = 0
        for chrom in sorted(chrs):
            for end, group in groupby(peaks[chrom], key=itemgetter("end")):
                n_peak += 1
                these_peaks = list(group)
                peak = these_peaks[0]
                peakname = "%s%d" % (peakprefix.decode(), n_peak)
                write("%s\t%d\t%d\t%d" % (chrom.decode(),
                                          peak['start']+1,
                                          peak['end'],
                                          peak['length']))
                write("\t%.6g" % (round(peak['pileup'], 2)))  # pileup height at summit
                write("\t%.6g" % (peak['pscore']))  # -log10pvalue at summit
                write("\t%.6g" % (peak['fc']))  # fold change at summit
                write("\t%.6g" % (peak['qscore']))  # -log10qvalue at summit
                write("\t%s" % peakname)
                write("\n")
        return



# cython: language_level=3
# cython: profile=True
# Time-stamp: <2024-10-15 11:25:35 Tao Liu>

"""Module Description: functions to find maxima minima or smooth the
signal tracks.

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""
# ------------------------------------
# Python modules
# ------------------------------------
from math import factorial as mathfactorial
from math import sqrt as mathsqrt

# ------------------------------------
# Other modules
# ------------------------------------
# smoothing function
import numpy as np
import cython
import cython.cimports.numpy as cnp
from cython.cimports.cpython import bool


@cython.ccall
def maxima(signal: cnp.ndarray(cython.float, ndim=1),
           window_size: cython.int = 51) -> cnp.ndarray:
    """return the local maxima in a signal after applying a 2nd order
    Savitsky-Golay (polynomial) filter using window_size specified
    """
    m: cnp.ndarray(cython.int, ndim=1)
    smoothed: cnp.ndarray(cython.double, ndim=1)
    sign: cnp.ndarray(cython.double, ndim=1)
    diff: cnp.ndarray(cython.double, ndim=1)

    window_size = window_size//2*2+1  # to make an odd number
    smoothed = savitzky_golay_order2_deriv1(signal, window_size).round(16)
    sign = np.sign(smoothed)
    diff = np.diff(sign)
    m = np.where(diff <= -1)[0].astype("i4")
    return m


@cython.cfunc
def internal_minima(signal: cnp.ndarray(cython.float, ndim=1),
                    maxima: cnp.ndarray(cython.int, ndim=1)) -> cnp.ndarray:
    ret: cnp.ndarray(cython.int, ndim=1)
    n: cython.int = maxima.shape[0]
    i: cython.int

    if n == 0 or n == 1:
        ret = np.ndarray(0, 'i4')
        return ret
    else:
        ret = np.zeros(n - 1, 'i4')
        pos1 = maxima[0]
        for i in range(n - 1):
            pos2 = maxima[i + 1]
            ret[i] = np.where(signal[pos1:pos2] == signal[pos1:pos2].min())[0][0] + pos1
            pos1 = pos2
        return ret


@cython.cfunc
@cython.inline
def sqrt(threshold: cython.float) -> cython.float:
    return mathsqrt(threshold)


@cython.ccall
def enforce_peakyness(signal: cnp.ndarray(cython.float, ndim=1),
                      maxima: cnp.ndarray(cython.int, ndim=1)):
    """requires peaks described by a signal and a set of points where
    the signal is at a maximum to meet a certain set of criteria

    maxima which do not meet the required criteria are discarded

    criteria:
        for each peak:

            calculate a threshold of the maximum of its adjacent two
            minima plus the sqrt of that value

            subtract the threshold from the region bounded by those minima

            clip that region if negative values occur inside it

            require it be > 50 bp in width -- controlled by is_valied_peak()

            require that it not be too flat (< 6 unique values) --
            controlled by is_valid_peak()

    """
    minima: cnp.ndarray(cython.int, ndim=1) = internal_minima(signal, maxima)
    new_signal: cnp.ndarray(cython.float, ndim=1)
    n: cython.int = minima.shape[0]
    threshold: cython.float
    peaky_maxima: cnp.ndarray(cython.int, ndim=1) = maxima.copy()
    j: cython.int = 0

    if n == 0:
        return maxima

    threshold = signal[minima[0]]
    threshold += sqrt(threshold)
    new_signal = signal[0:minima[0]] - threshold - sqrt(threshold)

    if is_valid_peak(new_signal, maxima[0]):
        peaky_maxima[0] = maxima[0]
        j += 1
    for i in range(n - 1):
        threshold = max(signal[minima[i]], signal[minima[i + 1]])
        threshold += sqrt(threshold)
        new_signal = signal[minima[i]:minima[i+1]] - threshold
        new_maximum = maxima[i+1] - minima[i]
        if is_valid_peak(new_signal, new_maximum):
            peaky_maxima[j] = maxima[i + 1]
            j += 1
    threshold = signal[minima[-1]]
    threshold += sqrt(threshold)
    new_signal = signal[minima[-1]:] - threshold
    new_maximum = maxima[-1] - minima[-1]
    if is_valid_peak(new_signal, new_maximum):
        peaky_maxima[j] = maxima[-1]
        j += 1
    peaky_maxima.resize(j, refcheck=False)
    return peaky_maxima


# hardcoded minimum peak width = 50
@cython.cfunc
def is_valid_peak(signal: cnp.ndarray(cython.float, ndim=1),
                  maximum: cython.int) -> bool:
    s: cnp.ndarray
    length: cython.int

    s = hard_clip(signal, maximum)
    length = s.shape[0]
    if length < 50:
        return False
    elif too_flat(s):
        return False
    return True


# require at least 6 different float values -- prevents broad flat peaks
@cython.cfunc
def too_flat(signal: cnp.ndarray(cython.float, ndim=1)) -> bool:
    """return whether signal has at least 6 unique values
    """
    return np.unique(signal).shape[0] < 6


# hard clip a region with negative values
@cython.cfunc
def hard_clip(signal: cnp.ndarray(cython.float, ndim=1),
              maximum: cython.int) -> cnp.ndarray:
    """clip the signal in both directions at the nearest values <= 0
    to position maximum
    """
    i: cython.int
    left: cython.int = 0
    right: cython.int = signal.shape[0]

    # clip left
    for i in range(right - maximum, 0):
        if signal[-i] < 0:
            left = i
            break
    for i in range(maximum, right):
        if signal[i] < 0:
            right = i
            break
    return signal[left:right]


@cython.ccall
def enforce_valleys(signal: cnp.ndarray(cython.float, ndim=1),
                    summits: cnp.ndarray(cython.int, ndim=1),
                    min_valley: cython.float = 0.8) -> cnp.ndarray:
    """require a value of <= min_valley * lower summit
    between each pair of summits
    """
    req_min: cython.float
    v: cython.float
    prev_v: cython.float

    summit_pos: cython.int
    prev_summit_pos: cython.int
    n_summits: cython.int
    n_valid_summits: cython.int

    valid_summits: cnp.ndarray(cython.int, ndim=1)

    n_summits = summits.shape[0]
    n_valid_summits = 1
    valid_summits = summits.copy()
    # Remove peaks that do not have sufficient valleys
    if n_summits == 1:
        return summits
    for i in range(1, n_summits):
        prev_summit_pos = valid_summits[n_valid_summits-1]
        summit_pos = summits[i]
        prev_v = signal[prev_summit_pos]
        v = signal[summit_pos]
        req_min = min_valley * min(prev_v, v)
        if (signal[prev_summit_pos:summit_pos] < req_min).any():
            valid_summits[n_valid_summits] = summit_pos
            n_valid_summits += 1
        elif v > prev_v:
            valid_summits[n_valid_summits-1] = summit_pos
    valid_summits.resize(n_valid_summits, refcheck=False)
    return valid_summits


# Modified from http://www.scipy.org/Cookbook/SavitzkyGolay
# positive window_size not enforced anymore
# needs sane input paramters, window size > 4
# switched to double precision for internal accuracy
@cython.ccall
def savitzky_golay_order2_deriv1(signal: cnp.ndarray(cython.float, ndim=1),
                                 window_size: cython.int) -> cnp.ndarray:
    """Smooth (and optionally differentiate) data with a Savitzky-Golay filter.
    The Savitzky-Golay filter removes high frequency noise from data.
    It has the advantage of preserving the original shape and
    features of the signal better than other types of filtering
    approaches, such as moving averages techhniques.
    Parameters
    ----------
    y : array_like, shape (N,)
        the values of the time history of the signal.
    window_size : int
        the length of the window. Must be an odd integer number.
    deriv: int
        the order of the derivative to compute (default = 0 means only smoothing)
    Returns
    -------
    ys : ndarray, shape (N)
        the smoothed signal (or it's n-th derivative).
    Notes
    -----
    The Savitzky-Golay is a type of low-pass filter, particularly
    suited for smoothing noisy data. The main idea behind this
    approach is to make for each point a least-square fit with a
    polynomial of high order over a odd-sized window centered at
    the point.

    References
    ----------
    .. [1] A. Savitzky, M. J. E. Golay, Smoothing and Differentiation of
       Data by Simplified Least Squares Procedures. Analytical
       Chemistry, 1964, 36 (8), pp 1627-1639.
    .. [2] Numerical Recipes 3rd Edition: The Art of Scientific Computing
       W.H. Press, S.A. Teukolsky, W.T. Vetterling, B.P. Flannery
       Cambridge University Press ISBN-13: 9780521880688
    """
    half_window: cython.int
    b: cnp.ndarray(cython.long, ndim=2)
    # pad the signal at the extremes with
    # values taken from the signal itself
    firstvals: cnp.ndarray(cython.float, ndim=1)
    lastvals: cnp.ndarray(cython.float, ndim=1)
    m: cnp.ndarray(cython.double, ndim=1)
    ret: cnp.ndarray(cython.double, ndim=1)

    if window_size % 2 != 1:
        window_size += 1
    half_window = (window_size - 1) // 2
    # precompute coefficients
    b = np.array([[1, k, k**2] for k in range(-half_window, half_window+1)],
                 dtype='i8')
    m = np.linalg.pinv(b)[1]
    # pad the signal at the extremes with
    # values taken from the signal itself
    firstvals = signal[0] - np.abs(signal[1:half_window+1][::-1] - signal[0])
    lastvals = signal[-1] + np.abs(signal[-half_window-1:-1][::-1] - signal[-1])
    signal = np.concatenate((firstvals, signal, lastvals))
    ret = np.convolve(m[::-1],
                      signal.astype("f8"),
                      mode='valid')
    return ret


# Another modified version from http://www.scipy.org/Cookbook/SavitzkyGolay
@cython.ccall
def savitzky_golay(y: cnp.ndarray(cython.float, ndim=1),
                   window_size: cython.int,
                   order: cython.int,
                   deriv: cython.int = 0,
                   rate: cython.int = 1) -> cnp.ndarray:
    """Smooth (and optionally differentiate) data with a Savitzky-Golay filter.
    The Savitzky-Golay filter removes high frequency noise from data.
    It has the advantage of preserving the original shape and
    features of the signal better than other types of filtering
    approaches, such as moving averages techniques.
    Parameters
    ----------
    y : array_like, shape (N,)
        the values of the time history of the signal.
    window_size : int
        the length of the window. Must be an odd integer number.
    order : int
        the order of the polynomial used in the filtering.
        Must be less then `window_size` - 1.
    deriv: int
        the order of the derivative to compute (default = 0 means only smoothing)
    Returns
    -------
    ys : ndarray, shape (N)
        the smoothed signal (or it's n-th derivative).
    Notes
    -----
    The Savitzky-Golay is a type of low-pass filter, particularly
    suited for smoothing noisy data. The main idea behind this
    approach is to make for each point a least-square fit with a
    polynomial of high order over a odd-sized window centered at
    the point.
    Examples
    --------
    t = np.linspace(-4, 4, 500)
    y = np.exp(-t**2) + np.random.normal(0, 0.05, t.shape)
    ysg = savitzky_golay(y, window_size=31, order=4)
    import matplotlib.pyplot as plt
    plt.plot(t, y, label='Noisy signal')
    plt.plot(t, np.exp(-t**2), 'k', lw=1.5, label='Original signal')
    plt.plot(t, ysg, 'r', label='Filtered signal')
    plt.legend()
    plt.show()
    References
    ----------
    .. [1] A. Savitzky, M. J. E. Golay, Smoothing and Differentiation of
       Data by Simplified Least Squares Procedures. Analytical
       Chemistry, 1964, 36 (8), pp 1627-1639.
    .. [2] Numerical Recipes 3rd Edition: The Art of Scientific Computing
       W.H. Press, S.A. Teukolsky, W.T. Vetterling, B.P. Flannery
       Cambridge University Press ISBN-13: 9780521880688
    """
    half_window: cython.int
    b: cnp.ndarray(cython.long, ndim=2)
    # pad the signal at the extremes with
    # values taken from the signal itself
    firstvals: cnp.ndarray(cython.float, ndim=1)
    lastvals: cnp.ndarray(cython.float, ndim=1)
    ret: cnp.ndarray(cython.float, ndim=1)
    m: cnp.ndarray(cython.double, ndim=1)

    try:
        window_size = np.abs(np.int(window_size))
        order = np.abs(np.int(order))
    except ValueError:
        raise ValueError("window_size and order have to be of type int")
    if window_size % 2 != 1 or window_size < 1:
        raise TypeError("window_size size must be a positive odd number")
    if window_size < order + 2:
        raise TypeError("window_size is too small for the polynomials order")
    half_window = (window_size - 1) // 2
    # precompute coefficients
    b = np.array([[k**i
                   for i in range(order + 1)]
                  for k in range(-half_window, half_window+1)])
    m = np.linalg.pinv(b)[deriv] * rate**deriv * mathfactorial(deriv)
    # pad the signal at the extremes with
    # values taken from the signal itself
    firstvals = y[0] - np.abs(y[1:half_window + 1][::-1] - y[0])
    lastvals = y[-1] + np.abs(y[-half_window - 1:-1][::-1] - y[-1])
    y = np.concatenate((firstvals, y, lastvals))
    ret = np.convolve(m[::-1], y, mode='valid').astype("float32")
    return ret



# cython: language_level=3
# cython: profile=True
# Time-stamp: <2024-10-24 15:20:26 Tao Liu>

"""Module Description: Detect peaks, main module

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""
# ------------------------------------
# Python modules
# ------------------------------------
import cython

# ------------------------------------
# MACS3 modules
# ------------------------------------
# from MACS3.Utilities.Constants import *
from MACS3.Signal.CallPeakUnit import CallerFromAlignments


@cython.cfunc
def subpeak_letters(i: cython.short) -> bytes:
    if i < 26:
        return chr(97+i).encode()
    else:
        return subpeak_letters(i // 26) + chr(97 + (i % 26)).encode()


class PeakDetect:
    """Class to do the peak calling.

    """
    def __init__(self,
                 opt=None,
                 treat=None,
                 control=None,
                 d=None,
                 maxgap=None,
                 minlen=None,
                 slocal=None,
                 llocal=None):
        """Initialize the PeakDetect object.

        """
        self.opt = opt
        self.info = opt.info
        self.debug = opt.debug
        self.warn = opt.warn

        self.treat = treat
        self.control = control
        self.ratio_treat2control = None
        self.peaks = None
        self.final_peaks = None
        self.PE_MODE = opt.PE_MODE
        self.scoretrack = None

        # self.femax = opt.femax
        # self.femin = opt.femin
        # self.festep = opt.festep

        self.log_pvalue = opt.log_pvalue    # -log10pvalue
        self.log_qvalue = opt.log_qvalue    # -log10qvalue
        if d is not None:
            self.d = d
        else:
            self.d = self.opt.d

        if opt.maxgap:
            self.maxgap = opt.maxgap
        else:
            self.maxgap = opt.tsize

        if opt.minlen:
            self.minlen = opt.minlen
        else:
            self.minlen = self.d

        self.end_shift = self.opt.shift
        self.gsize = opt.gsize

        self.nolambda = opt.nolambda

        if slocal is not None:
            self.sregion = slocal
        else:
            self.sregion = opt.smalllocal

        if llocal is not None:
            self.lregion = llocal
        else:
            self.lregion = opt.largelocal

        if (self.nolambda):
            self.info("#3 !!!! DYNAMIC LAMBDA IS DISABLED !!!!")
        # self.diag = opt.diag
        # self.save_score = opt.store_score
        # self.zwig_tr = opt.zwig_tr
        # self.zwig_ctl= opt.zwig_ctl

    def call_peaks(self):
        """Call peaks function.

        Scan the whole genome for peaks. RESULTS WILL BE SAVED IN
        self.final_peaks and self.final_negative_peaks.
        """
        if self.control:                # w/ control
            # if self.opt.broad:
            #    (self.peaks,self.broadpeaks) = self.__call_peaks_w_control()
            # else:
            self.peaks = self.__call_peaks_w_control()
        else:                           # w/o control
            # if self.opt.broad:
            #    (self.peaks,self.broadpeaks) = self.__call_peaks_wo_control()
            # else:
            self.peaks = self.__call_peaks_wo_control()
        return self.peaks

    def __call_peaks_w_control(self):
        """To call peaks with control data.

        A peak info type is a: dictionary

        key value: chromosome

        items: (peak start,peak end, peak length, peak summit, peak
        height, number of tags in peak region, peak pvalue, peak
        fold_enrichment) <-- tuple type

        While calculating pvalue:

        First, t and c will be adjusted by the ratio between total
        reads in treatment and total reads in control, depending on
        --to-small option.

        Then, t and c will be multiplied by the smallest peak size --
        self.d.

        Finally, a poisson CDF is applied to calculate one-side pvalue
        for enrichment.
        """
        lambda_bg: cython.float
        treat_scale: cython.float
        d: cython.float
        ctrl_scale_s: list
        ctrl_d_s: list
        control_total: cython.long
        # approx sum of treatment pileup values
        treat_sum: cython.long
        # approx sum of control pileup values
        control_sum: cython.long

        if self.PE_MODE:
            d = self.treat.average_template_length
            # in PE mode, entire fragment is counted as 1 in treatment
            # whereas both ends of fragment are counted in
            # control/input.
            control_total = self.control.total * 2
            treat_sum = self.treat.length
            control_sum = control_total * self.treat.average_template_length
            self.ratio_treat2control = float(treat_sum)/control_sum
        else:
            d = self.d
            control_total = self.control.total
            treat_sum = self.treat.total * self.d
            control_sum = self.control.total * self.d
            self.ratio_treat2control = float(treat_sum)/control_sum

        if self.opt.ratio != 1.0:
            self.ratio_treat2control = self.opt.ratio

        if self.opt.tocontrol:
            # if MACS decides to scale treatment to control data
            # because treatment is bigger
            lambda_bg = float(control_sum) / self.gsize
            treat_scale = 1/self.ratio_treat2control
        else:
            # if MACS decides to scale control to treatment because
            # control sample is bigger
            lambda_bg = float(treat_sum) / self.gsize
            treat_scale = 1.0

        # prepare d_s for control data
        if self.sregion:
            assert self.d <= self.sregion, f"{self.sregion:} can't be smaller than {self.d:}!"
        if self.lregion:
            assert self.d <= self.lregion, f"{self.lregion:} can't be smaller than {self.d:}!"
            assert self.sregion <= self.lregion, f"{self.lregion:} can't be smaller than {self.sregion:}!"

        # Now prepare a list of extension sizes
        ctrl_d_s = [self.d]   # note, d doesn't make sense in PE mode.
        # And a list of scaling factors for control
        ctrl_scale_s = []

        # d
        if not self.opt.tocontrol:
            # if user wants to scale everything to ChIP data
            tmp_v = self.ratio_treat2control
        else:
            tmp_v = 1.0
        ctrl_scale_s.append(tmp_v)

        # slocal size local
        if self.sregion:
            ctrl_d_s.append(self.sregion)
            if not self.opt.tocontrol:
                # if user want to scale everything to ChIP data
                tmp_v = float(self.d)/self.sregion*self.ratio_treat2control
            else:
                tmp_v = float(self.d)/self.sregion
            ctrl_scale_s.append(tmp_v)

        # llocal size local
        if self.lregion and self.lregion > self.sregion:
            ctrl_d_s.append(self.lregion)
            if not self.opt.tocontrol:
                # if user want to scale everything to ChIP data
                tmp_v = float(self.d)/self.lregion*self.ratio_treat2control
            else:
                tmp_v = float(self.d)/self.lregion
            ctrl_scale_s.append(tmp_v)

        # if self.PE_MODE:        # first d/scale are useless in PE mode
        #    ctrl_d_s = ctrl_d_s[1:]
        #    ctrl_scale_s = ctrl_scale_s[1:]
        #    print ctrl_d_s
        #    print ctrl_scale_s
        if self.nolambda:
            ctrl_d_s = []
            ctrl_scale_s = []

        scorecalculator = CallerFromAlignments(self.treat, self.control,
                                               d=d, ctrl_d_s=ctrl_d_s,
                                               treat_scaling_factor=treat_scale,
                                               ctrl_scaling_factor_s=ctrl_scale_s,
                                               end_shift=self.end_shift,
                                               lambda_bg=lambda_bg,
                                               save_bedGraph=self.opt.store_bdg,
                                               bedGraph_filename_prefix=self.opt.name,
                                               bedGraph_treat_filename=self.opt.bdg_treat,
                                               bedGraph_control_filename=self.opt.bdg_control,
                                               save_SPMR=self.opt.do_SPMR,
                                               cutoff_analysis_filename=self.opt.cutoff_analysis_file )

        if self.opt.trackline:
            scorecalculator.enable_trackline()

        # call peaks
        call_summits = self.opt.call_summits
        if call_summits:
            self.info("#3 Going to call summits inside each peak ...")

        if self.log_pvalue is not None:
            if self.opt.broad:
                self.info("#3 Call broad peaks with given level1 -log10pvalue cutoff and level2: %.5f, %.5f..." %
                          (self.log_pvalue, self.opt.log_broadcutoff))
                peaks = scorecalculator.call_broadpeaks(['p',],
                                                        lvl1_cutoff_s=[self.log_pvalue,],
                                                        lvl2_cutoff_s=[self.opt.log_broadcutoff,],
                                                        min_length=self.minlen,
                                                        lvl1_max_gap=self.maxgap,
                                                        lvl2_max_gap=self.maxgap*4,
                                                        cutoff_analysis=self.opt.cutoff_analysis)
            else:
                self.info("#3 Call peaks with given -log10pvalue cutoff: %.5f ..." % self.log_pvalue)
                peaks = scorecalculator.call_peaks(['p',], [self.log_pvalue,],
                                                   min_length=self.minlen,
                                                   max_gap=self.maxgap,
                                                   call_summits=call_summits,
                                                   cutoff_analysis=self.opt.cutoff_analysis)
        elif self.log_qvalue is not None:
            if self.opt.broad:
                self.info("#3 Call broad peaks with given level1 -log10qvalue cutoff and level2: %f, %f..." %
                          (self.log_qvalue, self.opt.log_broadcutoff))
                peaks = scorecalculator.call_broadpeaks(['q',],
                                                        lvl1_cutoff_s=[self.log_qvalue,],
                                                        lvl2_cutoff_s=[self.opt.log_broadcutoff,],
                                                        min_length=self.minlen,
                                                        lvl1_max_gap=self.maxgap,
                                                        lvl2_max_gap=self.maxgap*4,
                                                        cutoff_analysis=self.opt.cutoff_analysis)
            else:
                peaks = scorecalculator.call_peaks(['q',], [self.log_qvalue,],
                                                   min_length=self.minlen,
                                                   max_gap=self.maxgap,
                                                   call_summits=call_summits,
                                                   cutoff_analysis=self.opt.cutoff_analysis)
        scorecalculator.destroy()
        return peaks

    def __call_peaks_wo_control(self):
        """To call peaks without control data.

        A peak info type is a: dictionary

        key value: chromosome

        items: (peak start,peak end, peak length, peak summit, peak
        height, number of tags in peak region, peak pvalue, peak
        fold_enrichment) <-- tuple type

        While calculating pvalue:

        First, t and c will be adjusted by the ratio between total
        reads in treatment and total reads in control, depending on
        --to-small option.

        Then, t and c will be multiplied by the smallest peak size --
        self.d.

        Finally, a poisson CDF is applied to calculate one-side pvalue
        for enrichment.
        """
        lambda_bg: cython.float
        treat_scale: cython.float = 1
        d: cython.float
        ctrl_scale_s: list
        ctrl_d_s: list

        if self.PE_MODE:
            d = 0
        else:
            d = self.d
        treat_length = self.treat.length
        treat_total = self.treat.total

        # global lambda
        if self.PE_MODE:
        # this an estimator, we should maybe test it for accuracy?
            lambda_bg = treat_length / self.gsize
        else:
            lambda_bg = float(d) * treat_total / self.gsize
        treat_scale = 1.0

        # slocal and d-size local bias are not calculated!
        # nothing done here. should this match w control??

        if not self.nolambda:
            if self.PE_MODE:
                ctrl_scale_s = [float(treat_length) / (self.lregion*treat_total*2),]
            else:
                ctrl_scale_s = [float(self.d) / self.lregion,]
            ctrl_d_s = [self.lregion,]
        else:
            ctrl_scale_s = []
            ctrl_d_s = []

        scorecalculator = CallerFromAlignments(self.treat, None,
                                               d=d,
                                               ctrl_d_s=ctrl_d_s,
                                               treat_scaling_factor=treat_scale,
                                               ctrl_scaling_factor_s=ctrl_scale_s,
                                               end_shift=self.end_shift,
                                               lambda_bg=lambda_bg,
                                               save_bedGraph=self.opt.store_bdg,
                                               bedGraph_filename_prefix=self.opt.name,
                                               bedGraph_treat_filename=self.opt.bdg_treat,
                                               bedGraph_control_filename=self.opt.bdg_control,
                                               save_SPMR=self.opt.do_SPMR,
                                               cutoff_analysis_filename=self.opt.cutoff_analysis_file)

        if self.opt.trackline:
            scorecalculator.enable_trackline()

        # call peaks
        call_summits = self.opt.call_summits
        if call_summits:
            self.info("#3 Going to call summits inside each peak ...")

        if self.log_pvalue is not None:
            if self.opt.broad:
                self.info("#3 Call broad peaks with given level1 -log10pvalue cutoff and level2: %.5f, %.5f..." %
                          (self.log_pvalue, self.opt.log_broadcutoff))
                peaks = scorecalculator.call_broadpeaks(['p',],
                                                        lvl1_cutoff_s=[self.log_pvalue,],
                                                        lvl2_cutoff_s=[self.opt.log_broadcutoff,],
                                                        min_length=self.minlen,
                                                        lvl1_max_gap=self.maxgap,
                                                        lvl2_max_gap=self.maxgap*4,
                                                        cutoff_analysis=self.opt.cutoff_analysis)
            else:
                self.info("#3 Call peaks with given -log10pvalue cutoff: %.5f ..." % self.log_pvalue)
                peaks = scorecalculator.call_peaks(['p',], [self.log_pvalue,],
                                                   min_length=self.minlen,
                                                   max_gap=self.maxgap,
                                                   call_summits=call_summits,
                                                   cutoff_analysis=self.opt.cutoff_analysis)
        elif self.log_qvalue is not None:
            if self.opt.broad:
                self.info("#3 Call broad peaks with given level1 -log10qvalue cutoff and level2: %f, %f..." %
                          (self.log_qvalue, self.opt.log_broadcutoff))
                peaks = scorecalculator.call_broadpeaks(['q',],
                                                        lvl1_cutoff_s=[self.log_qvalue,],
                                                        lvl2_cutoff_s=[self.opt.log_broadcutoff,],
                                                        min_length=self.minlen,
                                                        lvl1_max_gap=self.maxgap,
                                                        lvl2_max_gap=self.maxgap*4,
                                                        cutoff_analysis=self.opt.cutoff_analysis)
            else:
                peaks = scorecalculator.call_peaks(['q',], [self.log_qvalue,],
                                                   min_length=self.minlen,
                                                   max_gap=self.maxgap,
                                                   call_summits=call_summits,
                                                   cutoff_analysis=self.opt.cutoff_analysis)
        scorecalculator.destroy()
        return peaks



# cython: language_level=3
# cython: profile=True
# Time-stamp: <2024-10-15 10:20:32 Tao Liu>
"""Module Description: Build shifting model

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""

# ------------------------------------
# Python modules
# ------------------------------------

# ------------------------------------
# MACS3 modules
# ------------------------------------
# from MACS3.Utilities.Constants import *
from MACS3.Signal.Pileup import naive_quick_pileup, naive_call_peaks

# ------------------------------------
# Other modules
# ------------------------------------
import cython
from cython.cimports.cpython import bool
import numpy as np
import cython.cimports.numpy as cnp

# ------------------------------------
# C lib
# ------------------------------------


class NotEnoughPairsException(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


@cython.cclass
class PeakModel:
    """Peak Model class.
    """
    # this can be PETrackI or FWTrack
    treatment: object
    # genome size
    gz: cython.double
    max_pairnum: cython.int
    umfold: cython.int
    lmfold: cython.int
    bw: cython.int
    d_min: cython.int
    tag_expansion_size: cython.int

    info: object
    debug: object
    warn: object
    error: object

    summary: str
    max_tags: cython.int
    peaksize: cython.int

    plus_line = cython.declare(cnp.ndarray, visibility="public")
    minus_line = cython.declare(cnp.ndarray, visibility="public")
    shifted_line = cython.declare(cnp.ndarray, visibility="public")
    xcorr = cython.declare(cnp.ndarray, visibility="public")
    ycorr = cython.declare(cnp.ndarray, visibility="public")

    d = cython.declare(cython.int, visibility="public")
    scan_window = cython.declare(cython.int, visibility="public")
    min_tags = cython.declare(cython.int, visibility="public")
    alternative_d = cython.declare(list, visibility="public")

    def __init__(self, opt, treatment, max_pairnum: cython.int = 500):
        # , double gz = 0, int umfold=30, int lmfold=10, int bw=200,
        # int ts = 25, int bg=0, bool quiet=False):
        self.treatment = treatment
        self.gz = opt.gsize
        self.umfold = opt.umfold
        self.lmfold = opt.lmfold
        # opt.tsize| test 10bps. The reason is that we want the best
        # 'lag' between left & right cutting sides. A tag will be
        # expanded to 10bps centered at cutting point.
        self.tag_expansion_size = 10
        # discard any predicted fragment sizes < d_min
        self.d_min = opt.d_min
        self.bw = opt.bw
        self.info = opt.info
        self.debug = opt.debug
        self.warn = opt.warn
        self.error = opt.warn
        self.max_pairnum = max_pairnum

    @cython.ccall
    def build(self):
        """Build the model. Main function of PeakModel class.

        1. prepare self.d, self.scan_window, self.plus_line,
        self.minus_line and self.shifted_line.

        2. find paired + and - strand peaks

        3. find the best d using x-correlation
        """
        paired_peakpos: dict
        num_paired_peakpos: cython.long
        c: bytes                # chromosome

        self.peaksize = 2*self.bw
        # mininum unique hits on single strand, decided by lmfold
        self.min_tags = int(round(float(self.treatment.total) *
                                  self.lmfold *
                                  self.peaksize / self.gz / 2))
        # maximum unique hits on single strand, decided by umfold
        self.max_tags = int(round(float(self.treatment.total) *
                                  self.umfold *
                                  self.peaksize / self.gz / 2))
        self.debug(f"#2 min_tags: {self.min_tags}; max_tags:{self.max_tags}; ")
        self.info("#2 looking for paired plus/minus strand peaks...")
        # find paired + and - strand peaks
        paired_peakpos = self.__find_paired_peaks()

        num_paired_peakpos = 0
        for c in list(paired_peakpos.keys()):
            num_paired_peakpos += len(paired_peakpos[c])

        self.info("#2 Total number of paired peaks: %d" % (num_paired_peakpos))

        if num_paired_peakpos < 100:
            self.error(f"#2 MACS3 needs at least 100 paired peaks at + and - strand to build the model, but can only find {num_paired_peakpos}! Please make your MFOLD range broader and try again. If MACS3 still can't build the model, we suggest to use --nomodel and --extsize 147 or other fixed number instead.")
            self.error("#2 Process for pairing-model is terminated!")
            raise NotEnoughPairsException("No enough pairs to build model")

        # build model, find the best d using cross-correlation
        self.__paired_peak_model(paired_peakpos)

    def __str__(self):
        """For debug...

        """
        return """
Summary of Peak Model:
  Baseline: %d
  Upperline: %d
  Fragment size: %d
  Scan window size: %d
""" % (self.min_tags, self.max_tags, self.d, self.scan_window)

    @cython.cfunc
    def __find_paired_peaks(self) -> dict:
        """Call paired peaks from fwtrackI object.

        Return paired peaks center positions.
        """
        i: cython.int
        chrs: list
        chrom: bytes
        plus_tags: cnp.ndarray(cython.int, ndim=1)
        minus_tags: cnp.ndarray(cython.int, ndim=1)
        plus_peaksinfo: list
        minus_peaksinfo: list
        paired_peaks_pos: dict  # return

        chrs = list(self.treatment.get_chr_names())
        chrs.sort()
        paired_peaks_pos = {}
        for i in range(len(chrs)):
            chrom = chrs[i]
            self.debug(f"Chromosome: {chrom}")
            # extract tag positions
            [plus_tags, minus_tags] = self.treatment.get_locations_by_chr(chrom)
            # look for + strand peaks
            plus_peaksinfo = self.__naive_find_peaks(plus_tags)
            self.debug("Number of unique tags on + strand: %d" % (plus_tags.shape[0]))
            self.debug("Number of peaks in + strand: %d" % (len(plus_peaksinfo)))
            if plus_peaksinfo:
                self.debug(f"plus peaks: first - {plus_peaksinfo[0]} ... last - {plus_peaksinfo[-1]}")
            # look for - strand peaks
            minus_peaksinfo = self.__naive_find_peaks(minus_tags)
            self.debug("Number of unique tags on - strand: %d" % (minus_tags.shape[0]))
            self.debug("Number of peaks in - strand: %d" % (len(minus_peaksinfo)))
            if minus_peaksinfo:
                self.debug(f"minus peaks: first - {minus_peaksinfo[0]} ... last - {minus_peaksinfo[-1]}")
            if not plus_peaksinfo or not minus_peaksinfo:
                self.debug("Chrom %s is discarded!" % (chrom))
                continue
            else:
                paired_peaks_pos[chrom] = self.__find_pair_center(plus_peaksinfo, minus_peaksinfo)
                self.debug("Number of paired peaks in this chromosome: %d" % (len(paired_peaks_pos[chrom])))
        return paired_peaks_pos

    @cython.cfunc
    def __naive_find_peaks(self,
                           taglist: cnp.ndarray(cython.int, ndim=1)) -> list:
        """Naively call peaks based on tags counting.

        Return peak positions and the tag number in peak region by a tuple list[(pos,num)].
        """
        peak_info: list
        pileup_array: list

        # store peak pos in every peak region and unique tag number in
        # every peak region
        peak_info = []

        # less than 2 tags, no need to call peaks, return []
        if taglist.shape[0] < 2:
            return peak_info

        # build pileup by extending both side to half peak size
        pileup_array = naive_quick_pileup(taglist, int(self.peaksize/2))
        peak_info = naive_call_peaks(pileup_array,
                                     self.min_tags,
                                     self.max_tags)

        return peak_info

    @cython.cfunc
    def __paired_peak_model(self, paired_peakpos: dict):
        """Use paired peak positions and treatment tag positions to
        build the model.

        Modify self.(d, model_shift size and scan_window size. and
        extra, plus_line, minus_line and shifted_line for plotting).

        """
        window_size: cython.int
        i: cython.int
        chroms: list
        paired_peakpos_chrom: object

        tags_plus: cnp.ndarray(cython.int, ndim=1)
        tags_minus: cnp.ndarray(cython.int, ndim=1)
        plus_start: cnp.ndarray(cython.int, ndim=1)
        plus_end: cnp.ndarray(cython.int, ndim=1)
        minus_start: cnp.ndarray(cython.int, ndim=1)
        minus_end: cnp.ndarray(cython.int, ndim=1)
        plus_line: cnp.ndarray(cython.int, ndim=1)
        minus_line: cnp.ndarray(cython.int, ndim=1)

        plus_data: cnp.ndarray
        minus_data: cnp.ndarray
        xcorr: cnp.ndarray
        ycorr: cnp.ndarray
        i_l_max: cnp.ndarray

        window_size = 1+2*self.peaksize+self.tag_expansion_size
        # for plus strand pileup
        self.plus_line = np.zeros(window_size, dtype="i4")
        # for minus strand pileup
        self.minus_line = np.zeros(window_size, dtype="i4")
        # for fast pileup
        plus_start = np.zeros(window_size, dtype="i4")
        # for fast pileup
        plus_end = np.zeros(window_size, dtype="i4")
        # for fast pileup
        minus_start = np.zeros(window_size, dtype="i4")
        # for fast pileup
        minus_end = np.zeros(window_size, dtype="i4")
        self.debug("start model_add_line...")
        chroms = list(paired_peakpos.keys())

        for i in range(len(chroms)):
            paired_peakpos_chrom = paired_peakpos[chroms[i]]
            (tags_plus, tags_minus) = self.treatment.get_locations_by_chr(chroms[i])
            # every paired peak has plus line and minus line
            self.__model_add_line(paired_peakpos_chrom,
                                  tags_plus,
                                  plus_start,
                                  plus_end)
            self.__model_add_line(paired_peakpos_chrom,
                                  tags_minus,
                                  minus_start,
                                  minus_end)

        self.__count(plus_start, plus_end, self.plus_line)
        self.__count(minus_start, minus_end, self.minus_line)

        self.debug("start X-correlation...")
        # Now I use cross-correlation to find the best d
        plus_line = self.plus_line
        minus_line = self.minus_line

        # normalize first
        minus_data = (minus_line - minus_line.mean())/(minus_line.std()*len(minus_line))
        plus_data = (plus_line - plus_line.mean())/(plus_line.std()*len(plus_line))

        # cross-correlation
        ycorr = np.correlate(minus_data, plus_data, mode="full")[window_size-self.peaksize:window_size+self.peaksize]
        xcorr = np.linspace(len(ycorr)//2*-1, len(ycorr)//2, num=len(ycorr))

        # smooth correlation values to get rid of local maximums from small fluctuations.
        # window size is by default 11.
        ycorr = smooth(ycorr, window="flat")

        # all local maximums could be alternative ds.
        i_l_max = np.r_[False, ycorr[1:] > ycorr[:-1]] & np.r_[ycorr[:-1] > ycorr[1:], False]
        i_l_max = np.where(i_l_max)[0]
        i_l_max = i_l_max[xcorr[i_l_max] > self.d_min]
        i_l_max = i_l_max[np.argsort(ycorr[i_l_max])[::-1]]

        self.alternative_d = sorted([int(x) for x in xcorr[i_l_max]])
        assert len(self.alternative_d) > 0, "No proper d can be found! Tweak --mfold?"

        self.d = xcorr[i_l_max[0]]

        self.ycorr = ycorr
        self.xcorr = xcorr

        self.scan_window = max(self.d, self.tag_expansion_size)*2

        self.info("#2 Model building with cross-correlation: Done")

        return True

    @cython.cfunc
    def __model_add_line(self,
                         pos1: list,
                         pos2: cnp.ndarray(cython.int, ndim=1),
                         start: cnp.ndarray(cython.int, ndim=1),
                         end: cnp.ndarray(cython.int, ndim=1)):
        """Project each pos in pos2 which is included in
        [pos1-self.peaksize,pos1+self.peaksize] to the line.

        pos1: paired centers -- list of coordinates
        pos2: tags of certain strand -- a numpy.array object
        line: numpy array object where we pileup tags

        """
        i1: cython.int
        i2: cython.int
        i2_prev: cython.int
        i1_max: cython.int
        i2_max: cython.int
        last_p2: cython.int
        psize_adjusted1: cython.int
        p1: cython.int
        p2: cython.int
        max_index: cython.int
        s: cython.int
        e: cython.int

        i1 = 0                  # index for pos1
        i2 = 0                  # index for pos2 index for pos2 in
        # previous pos1 [pos1-self.peaksize,pos1+self.peaksize] region
        i2_prev = 0
        i1_max = len(pos1)
        i2_max = pos2.shape[0]
        flag_find_overlap = False

        max_index = start.shape[0] - 1

        # half window
        psize_adjusted1 = self.peaksize + self.tag_expansion_size // 2

        while i1 < i1_max and i2 < i2_max:
            p1 = pos1[i1]
            p2 = pos2[i2]

            if p1-psize_adjusted1 > p2:
                # move pos2
                i2 += 1
            elif p1+psize_adjusted1 < p2:
                # move pos1
                i1 += 1
                i2 = i2_prev    # search minus peaks from previous index
                flag_find_overlap = False
            else:               # overlap!
                if not flag_find_overlap:
                    flag_find_overlap = True
                    # only the first index is recorded
                    i2_prev = i2
                # project
                s = max(int(p2-self.tag_expansion_size/2-p1+psize_adjusted1), 0)
                start[s] += 1
                e = min(int(p2+self.tag_expansion_size/2-p1+psize_adjusted1), max_index)
                end[e] -= 1
                i2 += 1
        return

    @cython.cfunc
    def __count(self,
                start: cnp.ndarray(cython.int, ndim=1),
                end: cnp.ndarray(cython.int, ndim=1),
                line: cnp.ndarray(cython.int, ndim=1)):
        """
        """
        i: cython.int
        pileup: cython.long

        pileup = 0
        for i in range(line.shape[0]):
            pileup += start[i] + end[i]
            line[i] = pileup
        return

    @cython.cfunc
    def __find_pair_center(self,
                           pluspeaks: list,
                           minuspeaks: list):
        # index for plus peaks
        ip: cython.long = 0
        # index for minus peaks
        im: cython.long = 0
        # index for minus peaks in previous plus peak
        im_prev: cython.long = 0
        pair_centers: list
        ip_max: cython.long
        im_max: cython.long
        flag_find_overlap: bool
        pp: cython.int
        mp: cython.int
        pn: cython.float
        mn: cython.float

        pair_centers = []
        ip_max = len(pluspeaks)
        im_max = len(minuspeaks)
        self.debug(f"ip_max: {ip_max}; im_max: {im_max}")
        flag_find_overlap = False
        while ip < ip_max and im < im_max:
            # for (peakposition, tagnumber in peak)
            (pp, pn) = pluspeaks[ip]
            (mp, mn) = minuspeaks[im]
            if pp-self.peaksize > mp:
                # move minus
                im += 1
            elif pp+self.peaksize < mp:
                # move plus
                ip += 1
                im = im_prev    # search minus peaks from previous index
                flag_find_overlap = False
            else:               # overlap!
                if not flag_find_overlap:
                    flag_find_overlap = True
                    # only the first index is recorded
                    im_prev = im
                # number tags in plus and minus peak region are comparable...
                if pn/mn < 2 and pn/mn > 0.5:
                    if pp < mp:
                        pair_centers.append((pp+mp)//2)
                im += 1
        if pair_centers:
            self.debug(f"Paired centers: first - {pair_centers[0]} ... second - {pair_centers[-1]} ")
        return pair_centers


# smooth function from SciPy cookbook:
# http://www.scipy.org/Cookbook/SignalSmooth
@cython.ccall
def smooth(x,
           window_len: cython.int = 11,
           window: str = 'hanning'):
    """smooth the data using a window with requested size.

    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal
    (with the window size) in both ends so that transient parts are minimized
    in the beginning and end part of the output signal.

    input:
        x: the input signal
        window_len: the dimension of the smoothing window; should be
                    an odd integer
        window: the type of window from 'flat', 'hanning', 'hamming',
                'bartlett', 'blackman' flat window will produce a
                moving average smoothing.

    output:
        the smoothed signal

    example:

    t=linspace(-2,2,0.1)
    x=sin(t)+randn(len(t))*0.1
    y=smooth(x)

    see also:

    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman,
    numpy.convolve scipy.signal.lfilter

    TODO: the window parameter could be the window itself if an array
          instead of a string

    NOTE: length(output) != length(input), to correct this: return
          y[(window_len/2-1):-(window_len/2)] instead of just y.
    """

    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")

    if window_len < 3:
        return x

    if window not in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError("Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'")

    s = np.r_[x[window_len-1:0:-1], x, x[-1:-window_len:-1]]

    if window == 'flat':        # moving average
        w = np.ones(window_len, 'd')
    else:
        w = eval('np.'+window+'(window_len)')

    y = np.convolve(w/w.sum(), s, mode='valid')
    return y[(window_len//2):-(window_len//2)]



# cython: language_level=3
# cython: profile=True
# Time-stamp: <2025-11-14 16:55:58 Tao Liu>

"""Module for Region classe.

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""

# ------------------------------------
# python modules
# ------------------------------------
# import random
# import re
# import sys

# ------------------------------------
# MACS3 modules
# ------------------------------------

# from MACS3.Utilities.Constants import *

# ------------------------------------
# Other modules
# ------------------------------------
import cython
from cython.cimports.cpython import bool

# ------------------------------------
# constants
# ------------------------------------
__version__ = "Region $Revision$"
__author__ = "Tao Liu <vladimir.liu@gmail.com>"
__doc__ = "Region class"

# ------------------------------------
# Misc functions
# ------------------------------------

# ------------------------------------
# Classes
# ------------------------------------


@cython.cclass
class Regions:
    """Container for genomic regions organized by chromosome.
    
    This class stores genomic regions (intervals) organized by chromosome
    and provides operations for region manipulation including sorting,
    merging overlaps, expanding, and set operations like intersection
    and exclusion.
    
    Attributes:
        regions (dict): Dictionary mapping chromosome names (bytes) to 
            lists of region tuples (start, end). Regions are 0-based,
            half-open intervals [start, end).
        total (int): Total count of regions across all chromosomes.
    """
    regions = cython.declare(dict, visibility='public')
    total = cython.declare(cython.int, visibility='public')
    __sorted: bool
    __merged: bool

    def __init__(self):
        self.regions = {}
        self.__sorted = False
        self.__merged = False
        self.total = 0

    def __getitem__(self, chrom):
        return self.regions[chrom]

    @cython.ccall
    def pop(self, n: cython.int):
        """Remove and return the first n regions.
        
        Removes the first n regions from the Regions object across 
        chromosomes in sorted order and returns them as a new Regions 
        object. The current object is modified.
        
        Args:
            n (int): Number of regions to pop.
            
        Returns:
            Regions: New Regions object containing the first n regions.
            
        Examples:
            .. code-block:: python

                regions.add_loc(b'chr1', 1000, 2000)
                regions.add_loc(b'chr1', 3000, 4000)
                regions.add_loc(b'chr2', 5000, 6000)
                regions.sort()
                first_two = regions.pop(2)
                print(first_two.total)  # Output: 2
                print(regions.total)    # Output: 1
        """
        # when called, pop the first n regions in Regions class. Self
        # will be modified.
        clist: list             # for chromosomes
        tmp_l: cython.int
        chrom: bytes
        n_taken: cython.int     # remember the number of regions prepared
        ret: object             # returned Regions

        if self.total == 0:
            raise Exception("None left")

        clist = sorted(list(self.regions.keys()))
        n_taken = n
        ret = Regions()
        for chrom in sorted(clist):
            ret.regions[chrom] = self.regions[chrom][:n_taken]
            self.regions[chrom] = self.regions[chrom][n_taken:]
            if not self.regions[chrom]:
                # remove this chromosome if there is none left
                self.regions.pop(chrom)
            tmp_l = len(ret.regions[chrom])
            ret.total += tmp_l
            # calculate remained
            self.total -= tmp_l
            # print(ret.total, self.total)
            n_taken -= tmp_l
            if not n_taken:
                # when there is no need, quit loop
                break
        return ret

    @cython.ccall
    def init_from_PeakIO(self, peaks):
        """Initialize the object with a PeakIO object.

        Note: I intentionally forgot to check if peakio is actually a
        PeakIO...
        """
        chrom: bytes
        ps: list
        p: object
        i: int

        peaks.sort()
        self.total = 0
        for chrom in sorted(peaks.get_chr_names()):
            ps = peaks.get_data_from_chrom(chrom)
            self.regions[chrom] = []
            for i in range(len(ps)):
                p = ps[i]
                self.regions[chrom].append((p['start'], p['end']))
                self.total += 1
        self.sort()

    @cython.ccall
    def add_loc(self, chrom: bytes, start: int, end: int):
        """Add a region to a specific chromosome.
        
        Adds a single region interval to the specified chromosome. The 
        region is added without immediate sorting. Call sort() to 
        maintain order.
        
        Args:
            chrom (bytes): Chromosome name.
            start (int): Start position (0-based, inclusive).
            end (int): End position (0-based, exclusive).
            
        Returns:
            None
        """
        if self.regions.has_key(chrom):
            self.regions[chrom].append((start, end))
        else:
            self.regions[chrom] = [(start, end),]
        self.total += 1
        self.__sorted = False
        self.__merged = False
        return

    @cython.ccall
    def sort(self):
        """Sort all regions within each chromosome by start position.
        
        Sorts regions for each chromosome. This method is idempotent—
        calling it multiple times has the same effect as calling once.
        
        Returns:
            None
        """
        chrom: bytes

        if self.__sorted:
            return
        for chrom in sorted(self.regions.keys()):
            self.regions[chrom].sort()
        self.__sorted = True

    @cython.ccall
    def total_length(self) -> cython.long:
        """Calculate the total length covered by all regions.
        
        Returns the sum of lengths of all regions across all chromosomes.
        Overlapping regions are first merged to avoid double-counting.
        
        Returns:
            int: Total nucleotide length covered by all regions.
            
        """
        chrom: bytes
        ps: list
        i: cython.int
        tl: cython.int
        s: cython.int
        e: cython.int
        self.merge_overlap()
        tl = 0
        for chrom in sorted(self.regions.keys()):
            ps = self.regions[chrom]
            for i in range(len(ps)):
                s, e = ps[i]
                tl += e - s
        return tl

    @cython.ccall
    def get_chr_names(self) -> set:
        """Get all chromosome names present in the Regions object.
        
        Returns:
            set: Set of chromosome names (bytes) present in regions.
            
        Examples:
            .. code-block:: python

                from MACS3.Signal.Region import Regions
                regions = Regions()
                regions.add_loc(b'chr1', 1000, 2000)
                regions.add_loc(b'chr1', 3000, 4000)
                regions.add_loc(b'chr2', 5000, 6000)
                print(regions.get_chr_names())
                # Output: {b'chr1', b'chr2'}
        """
        return set(sorted(self.regions.keys()))

    @cython.ccall
    def expand(self, flanking: int):
        """Expand all regions by a fixed distance in both directions.
        
        Extends the start and end positions of each region by 'flanking'
        base pairs. Start positions are expanded leftward but capped at 0.
        The regions are automatically re-sorted after expansion.
        
        Args:
            flanking (int): Number of base pairs to expand in each direction.
            
        Returns:
            None: Modifies the object in place.
            
        Examples:
            .. code-block:: python

                regions.expand(100)
                print(regions.regions[b'chr1'])  # Output: [(900, 2100)]
                
                # With capping at 0 for start positions
                regions2 = Regions()
                regions2.add_loc(b'chr1', 50, 150)
                regions2.expand(100)
                print(regions2.regions[b'chr1'])
                # Output: [(0, 250)]  # Start capped at 0
        """
        chrom: bytes
        ps: list
        i: cython.int

        self.sort()
        for chrom in sorted(self.regions.keys()):
            ps = self.regions[chrom]
            for i in range(len(ps)):
                ps[i] = (max(0, ps[i][0] - flanking), ps[i][1] + flanking)
            ps.sort()
            self.regions[chrom] = ps
        self.__merged = False

    @cython.ccall
    def merge_overlap(self):
        """Merge overlapping or adjacent regions within each chromosome.
        
        Combines regions that overlap or are adjacent. After merging,
        regions are re-sorted. This operation is idempotent and modifies
        the object in place. The total count is updated.
        
        Returns:
            bool: True if merge was performed, False if already merged.
            
        Examples:
            .. code-block:: python
            
                regions = Regions()
                regions.add_loc(b'chr1', 1000, 2000)
                regions.add_loc(b'chr1', 1500, 3000)  # Overlaps with first
                result = regions.merge_overlap()
                print(regions.regions[b'chr1'])  # Output: [(1000, 3000)]
        """
        chrom: bytes
        s_new_region: cython.int
        e_new_region: cython.int
        i: cython.int
        regions: dict
        new_regions: dict
        chrs: list
        regions_chr: list
        prev_region: tuple

        if self.__merged:
            return
        self.sort()
        regions = self.regions
        new_regions = {}

        chrs = list(regions.keys())
        chrs.sort()
        self.total = 0
        for i in range(len(chrs)):
            chrom = chrs[i]
            new_regions[chrom] = []
            n_append = new_regions[chrom].append
            prev_region = ()
            regions_chr = regions[chrom]
            for i in range(len(regions_chr)):
                if not prev_region:
                    prev_region = regions_chr[i]
                    continue
                else:
                    if regions_chr[i][0] <= prev_region[1]:
                        s_new_region = prev_region[0]
                        e_new_region = regions_chr[i][1]
                        prev_region = (s_new_region, e_new_region)
                    else:
                        n_append(prev_region)
                        prev_region = regions_chr[i]
            if prev_region:
                n_append(prev_region)
            self.total += len(new_regions[chrom])
        self.regions = new_regions
        self.sort()
        self.__merged = True
        return True

    @cython.ccall
    def write_to_bed(self, fhd):
        """Write regions to a BED format file.
        
        Args:
            fhd: File handle opened for writing.
            
        Examples:
            .. code-block:: python
            
                with open('output.bed', 'w') as f:
                    regions.write_to_bed(f)
        """
        i: cython.int
        chrom: bytes
        region: tuple

        chrs = list(self.regions.keys())
        chrs.sort()
        for i in range(len(chrs)):
            chrom = chrs[i]
            for region in self.regions[chrom]:
                fhd.write("%s\t%d\t%d\n" % (chrom.decode(),
                                            region[0], region[1]))

    def __str__(self):
        i: cython.int
        chrom: bytes
        region: tuple
        ret: str

        ret = ""
        chrs = list(self.regions.keys())
        chrs.sort()
        for i in range(len(chrs)):
            chrom = chrs[i]
            for region in self.regions[chrom]:
                ret += "%s\t%d\t%d\n" % (chrom.decode(),
                                         region[0], region[1])
        return ret

    # cpdef object randomly_pick (self, int n, int seed = 12345):
    #     """Shuffle the regions and get n regions out of it. Return a
    #     new Regions object.
    #     """

    #     cdef:
    #         list all_pc
    #         list chrs
    #         bytes chrom
    #         object ret_peakio, p
    #     assert n > 0
    #     chrs = sorted(list(self.peaks.keys()))
    #     all_pc = []
    #     for chrom in chrs:
    #         all_pc.extend(self.peaks[chrom])
    #     random.seed(seed)
    #     all_pc = random.shuffle(all_pc)[:n]
    #     ret_peakio = PeakIO()
    #     for p in all_pc:
    #         ret_peakio.add_PeakContent (p["chrom"], p)
    #     return ret_peakio

    @cython.ccall
    def intersect(self, regions_object2):
        """Get the intersection with another Regions object.
        
        Returns a new Regions object containing only the regions that
        overlap between this object and regions_object2. For chromosomes
        present only in this object, all regions are included unchanged.
        
        Args:
            regions_object2 (Regions): Another Regions object to intersect with.
            
        Returns:
            Regions: New object containing intersecting regions.
            
        Examples:
            .. code-block:: python
            
                r1 = Regions()
                r1.add_loc(b'chr1', 1000, 3000)
                
                r2 = Regions()
                r2.add_loc(b'chr1', 2000, 4000)
                
                result = r1.intersect(r2)
                print(result.regions[b'chr1'])  # Output: [(2000, 3000)]
        """
        ret_regions_object: object
        regions1: dict
        regions2: dict
        chrs1: list
        chrs2: list
        four_coords: list
        k: bytes
        ret_regions: dict
        r1: tuple
        r2: tuple
        n_rl1: cython.long
        n_rl2: cython.long

        self.sort()
        regions1 = self.regions
        assert isinstance(regions_object2, Regions)

        regions_object2.sort()
        regions2 = regions_object2.regions

        ret_regions_object = Regions()
        ret_regions = dict()
        ret_regions_object.total = 0
        chrs1 = list(regions1.keys())
        chrs2 = list(regions2.keys())
        for k in chrs1:
            # print(f"chromosome {k}")
            if not chrs2.count(k):
                # no such chromosome in peaks1, then don't touch the
                # peaks in this chromosome
                ret_regions[k] = regions1[k]
                ret_regions_object.total += len(ret_regions[k])
                continue
            ret_regions[k] = []
            n_rl1 = len(regions1[k])    # number of remaining elements in regions1[k]
            n_rl2 = len(regions2[k])    # number of remaining elements in regions2[k]
            rl1_k = iter(regions1[k]).__next__
            rl2_k = iter(regions2[k]).__next__
            r1 = rl1_k()
            n_rl1 -= 1
            r2 = rl2_k()
            n_rl2 -= 1
            while (True):
                # we do this until there is no r1 or r2 left.
                if r2[0] < r1[1] and r1[0] < r2[1]:
                    # We found an overlap, now get the intersecting
                    # region.
                    four_coords = sorted([r1[0], r1[1], r2[0], r2[1]])
                    ret_regions[k].append((four_coords[1], four_coords[2]))
                if r1[1] < r2[1]:
                    # in this case, we need to move to the next r1,
                    if n_rl1:
                        r1 = rl1_k()
                        n_rl1 -= 1
                    else:
                        # no more r1 left
                        break
                else:
                    # in this case, we need to move the next r2
                    if n_rl2:
                        r2 = rl2_k()
                        n_rl2 -= 1
                    else:
                        # no more r2 left
                        break
            ret_regions_object.total += len(ret_regions[k])

        ret_regions_object.regions = ret_regions
        ret_regions_object.sort()
        return ret_regions_object

    @cython.ccall
    def exclude(self, regions_object2):
        """Remove regions that overlap with another Regions object.
        
        Removes all regions from this object that overlap with regions
        in regions_object2. Regions present only in this object are kept.
        The object is modified in place.
        
        Args:
            regions_object2 (Regions): Another Regions object defining
                regions to exclude.
            
        Returns:
            None: Modifies the object in place.
            
        Examples:
            .. code-block:: python
                
                from MACS3.Signal.Region import Regions
                r1 = Regions()
                r1.add_loc(b'chr1', 1000, 3000)
                r1.add_loc(b'chr1', 5000, 7000)
                
                r2 = Regions()
                r2.add_loc(b'chr1', 2000, 6000)  # Overlaps both regions
                
                r1.exclude(r2)
                print(r1.regions[b'chr1'])
                # Output: [(1000, 2000), (6000, 7000)]
        """
        regions1: dict
        regions2: dict
        chrs1: list
        chrs2: list
        k: bytes
        ret_regions: dict
        overlap_found: bool
        r1: tuple
        r2: tuple
        n_rl1: cython.long
        n_rl2: cython.long

        self.sort()
        regions1 = self.regions
        self.total = 0
        assert isinstance(regions_object2, Regions)
        regions_object2.sort()
        regions2 = regions_object2.regions

        ret_regions = dict()
        chrs1 = list(regions1.keys())
        chrs2 = list(regions2.keys())
        for k in chrs1:
            # print(f"chromosome {k}")
            if not chrs2.count(k):
                # no such chromosome in peaks1, then don't touch the peaks in this chromosome
                ret_regions[k] = regions1[k]
                self.total += len(ret_regions[k])
                continue
            ret_regions[k] = []
            n_rl1 = len(regions1[k])
            n_rl2 = len(regions2[k])
            rl1_k = iter(regions1[k]).__next__
            rl2_k = iter(regions2[k]).__next__
            overlap_found = False
            r1 = rl1_k()
            n_rl1 -= 1
            r2 = rl2_k()
            n_rl2 -= 1
            while (True):
                # we do this until there is no r1 or r2 left.
                if r2[0] < r1[1] and r1[0] < r2[1]:
                    # since we found an overlap, r1 will be skipped/excluded
                    # and move to the next r1
                    overlap_found = True
                    n_rl1 -= 1
                    if n_rl1 >= 0:
                        r1 = rl1_k()
                        overlap_found = False
                        continue
                    else:
                        break
                if r1[1] < r2[1]:
                    # in this case, we need to move to the next r1,
                    # we will check if overlap_found is true, if not, we put r1 in a new dict
                    if not overlap_found:
                        ret_regions[k].append(r1)
                    n_rl1 -= 1
                    if n_rl1 >= 0:
                        r1 = rl1_k()
                        overlap_found = False
                    else:
                        # no more r1 left
                        break
                else:
                    # in this case, we need to move the next r2
                    if n_rl2:
                        r2 = rl2_k()
                        n_rl2 -= 1
                    else:
                        # no more r2 left
                        break
            if n_rl1 >= 0:
                ret_regions[k].extend(regions1[k][-n_rl1-1:])
            self.total += len(ret_regions[k])

        self.regions = ret_regions
        self.__sorted = False
        self.sort()
        return



MACS_VERSION = "3.0.4"
MAX_PAIRNUM = 1000
MAX_LAMBDA = 100000
FESTEP = 20
BUFFER_SIZE = 100000  # np array will increase at step of 1 million items
READ_BUFFER_SIZE = 10000000     # 10M bytes for read buffer size
N_MP = 2                        # Number of processers

# Effective genome size, collected from
# https://deeptools.readthedocs.io/en/develop/content/feature/effectiveGenomeSize.html
EFFECTIVEGS = {"hs": 2913022398,                  # GRCh38
               "mm": 2652783500,                  # GRCm38
               "ce": 100286401,                   # WBcel235
               "dm": 142573017}                   # dm6
